# -*- coding: utf-8 -*-

# Copyright (c) 2019 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import re

import numpy as np
from matplotlib.path import Path
from scipy.spatial import cKDTree


class SpaceMapper(object):
    """translate between world coordinates and normalized UV space"""

    def __init__(self, dfile, ptres=1.0, rotation=0.0, tolerance=1.0):
        #: float: ccw rotation (in degrees) for point grid on plane
        self.rotation = rotation
        #: float: tolerance for point search when using point list for area
        self.tolerance = tolerance
        #: float: point resolution for area
        self.ptres = ptres
        self.pt_kd = None
        #: np.array: boundary frame for translating between coordinates
        #: [[xmin ymin zmin] [xmax ymax zmax]]
        self.bbox = dfile

    @property
    def pt_kd(self):
        """point kdtree for spatial queries built at first use"""
        if self._pt_kd is None:
            self._pt_kd = cKDTree(self.pts())
        return self._pt_kd

    @pt_kd.setter
    def pt_kd(self, pt_kd):
        self._pt_kd = pt_kd

    @property
    def bbox(self):
        """np.array of shape (3,2): bounding box"""
        return self._bbox

    @property
    def sf(self):
        """bbox scale factor"""
        return self._sf

    @property
    def ptshape(self):
        """shape of point grid"""
        return self._ptshape

    @property
    def npts(self):
        """number of points"""
        return int(np.product(self.ptshape))

    @bbox.setter
    def bbox(self, plane):
        """read radiance geometry file as boundary path"""
        paths, bbox = self._rad_scene_to_bbox(plane)
        self._bbox = bbox
        self._sf = self.bbox[1, 0:2] - self.bbox[0, 0:2]
        self._path = []
        for pt in paths:
            p = (np.concatenate((pt, [pt[0]])) - bbox[0, 0:2])/self._sf
            xy = Path(p, closed=True)
            self._path.append(xy)
        size = (bbox[1, 0:2] - bbox[0, 0:2])/self.ptres
        self._ptshape = np.ceil(size).astype(int)

    def _ro_pts(self, points, rdir=-1):
        """
        rotate points

        Parameters
        ----------
        points: np.ndarray
            world coordinate points of shape (N, 3)
        rdir: -1 or 1
            rotation direction:
                -1 to rotate from uv space
                1 to rotate to uvspace

        Returns
        -------
        """
        cos = np.cos(rdir * self.rotation * np.pi / 180)
        sin = np.sin(rdir * self.rotation * np.pi / 180)
        rmtx = np.array([[cos, sin, 0], [-sin, cos, 0], [0, 0, 1]])
        rm = np.matmul(rmtx, points.T)
        return rm.T

    def uv2pt(self, uv):
        """convert UV --> world

        Parameters
        ----------
        uv: np.array
            normalized UV coordinates of shape (N, 2)

        Returns
        -------
        pt: np.array
            world xyz coordinates of shape (N, 3)
        """
        sf = self.bbox[1] - self.bbox[0]
        uv = np.hstack((uv, np.zeros(len(uv)).reshape(-1, 1)))
        pt = self.bbox[0] + uv * sf
        return self._ro_pts(pt)

    def pt2uv(self, xyz):
        """convert world --> UV

        Parameters
        ----------
        xyz: np.array
            world xyz coordinates, shape (N, 3)

        Returns
        -------
        uv: np.array
            normalized UV coordinates of shape (N, 2)
        """
        uv = (self._ro_pts(xyz, rdir=1) - self.bbox[0])[:, 0:2] / self._sf
        return uv

    def idx2pt(self, idx):
        shape = self.ptshape
        si = np.stack(np.unravel_index(idx, shape)).T
        return self.uv2pt((si + .5)/shape)

    def pts(self):
        shape = self.ptshape
        return self.idx2pt(np.arange(np.product(shape)))

    def in_area(self, xyz):
        """check if point is in boundary path

        Parameters
        ----------
        xyz: np.array
            uv coordinates, shape (N, 3)

        Returns
        -------
        mask: np.array
            boolean array, shape (N,)
        """
        uv = self.pt2uv(xyz)
        path = self._path
        if path is None:
            return np.full((uv.shape[0]), True)
        else:
            result = np.empty((len(path), uv.shape[0]), bool)
            for i, p in enumerate(path):
                result[i] = p.contains_points(uv)
        return np.any(result, 0)

    def _rad_scene_to_bbox(self, plane):
        with open(plane, 'r') as f:
            dl = [i for i in re.split(r'[\n\r]+', f.read().strip())
                  if not bool(re.match(r'#', i))]
            rad = " ".join(dl).split()
        pgs = [i for i, x in enumerate(rad) if x == "polygon"]
        z = -1e10
        paths = []
        for pg in pgs:
            npt = int(int(rad[pg + 4])/3)
            pt = np.reshape([i for i in rad[pg + 5:pg + 5 + npt*3]],
                            (npt, 3)).astype(float)
            z = max(z, max(pt[:, 2]))
            pt2 = self._ro_pts(pt, rdir=1)
            paths.append(pt2[:, 0:2])
        bbox = np.array(paths)
        bbox = np.squeeze([np.amin(bbox, 1), np.amax(bbox, 1)])
        bbox = np.hstack([bbox, [[z], [z]]])
        return paths, bbox



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

import os
import pickle
import itertools
from concurrent.futures import ThreadPoolExecutor, as_completed

import numpy as np

from raytraverse import translate, plot
from raytraverse.lightfield.lightfield import LightField


class SunViewField(LightField):
    """container for sun view data

    Parameters
    ----------
    scene: raytraverse.scene.Scene
        scene class containing geometry, location and analysis plane
    suns: raytraverse.sunsetter.SunSetter
        prefix of data files to integrate
    rebuild: bool, optional
        build kd-tree even if one exists
    """

    def __init__(self, scene, suns, rebuild=False, rmraw=False):
        #: raytraverse.sunsetter.SunSetter
        self.suns = suns
        #: raytraverse.sunmapper.SunMapper
        self.sunmap = suns.map
        super().__init__(scene, rebuild=rebuild, prefix='sunview', rmraw=rmraw)

    def raw_files(self):
        """get list of files used to build field"""
        dfile = f'{self.scene.outdir}/{self.prefix}_vals.out'
        vfile = f'{self.scene.outdir}/{self.prefix}_vecs.out'
        final = f'{self.scene.outdir}/{self.prefix}_result.pickle'
        return [dfile, vfile, final]

    @property
    def raster(self):
        """individual pixels forming shape of the sun, stored as uv coorinates
        with basis viewmapper about sun direction and diameter. indexed like
        vec, lum, and omega
        """
        return self._raster

    @property
    def scene(self):
        """scene information

        :getter: Returns this integrator's scene
        :setter: Set this integrator's scene
        :type: raytraverse.scene.Scene
        """
        return self._scene

    @scene.setter
    def scene(self, scene):
        """Set this integrator's scene and load samples"""
        self._scene = scene
        kdfile = f'{self.scene.outdir}/{self.prefix}_kd_data.pickle'
        if os.path.isfile(kdfile) and not self.rebuild:
            f = open(kdfile, 'rb')
            self._vec = pickle.load(f)
            self._lum = pickle.load(f)
            self._omega = pickle.load(f)
            self._raster = pickle.load(f)
            f.close()
        else:
            dfile = f'{self.scene.outdir}/{self.prefix}_result.pickle'
            if not os.path.isfile(dfile):
                raise FileNotFoundError("No results files found, have you run"
                                        f" a Sampler of type {self.prefix} for"
                                        f" scene {self.scene.outdir}?")
            f = open(dfile, 'rb')
            vls = [pickle.load(f) for i in range(3)]
            f.close()
            (self._vec, self._lum,
             self._omega, self._raster) = self._build_suns(*vls)
            f = open(kdfile, 'wb')
            pickle.dump(self.vec, f, protocol=4)
            pickle.dump(self.lum, f, protocol=4)
            pickle.dump(self.omega, f, protocol=4)
            pickle.dump(self.raster, f, protocol=4)
            f.close()

    def items(self):
        return self.vec.keys()

    def _grp_by_sun(self, vecs, lums, shape, pti, suni):
        scalefac = ((self.sunmap.viewangle/2*np.pi/180)**2)
        omega0 = scalefac*np.pi/np.prod(shape)
        xyz = self.sunmap.uv2xyz(vecs, i=suni)
        v = np.mean(xyz, 0)
        lm = np.mean(lums)
        og = omega0*len(vecs)
        return pti, suni, v, lm, og, vecs

    def _build_suns(self, vecs, lums, shape):
        """loop through points/suns and group adjacent rays"""
        vec = {}
        lum = {}
        omega = {}
        raster = {}
        futures = []
        with ThreadPoolExecutor() as exc:
            items = itertools.product(super().items(),
                                      range(self.suns.suns.shape[0]))
            for i, j in items:
                if len(vecs[i][j]) > 0:
                    futures.append(exc.submit(self._grp_by_sun, vecs[i][j],
                                              lums[i][j], shape, i, j))
            for fu in as_completed(futures):
                i, j, v, lm, og, ras = fu.result()
                vec[(i, j)] = v
                lum[(i, j)] = lm
                omega[(i, j)] = og
                raster[(i, j)] = ras
        return vec, lum, omega, raster

    def _to_pix(self, rxyz, atv, vm, res):
        if atv > 90:
            ppix = vm.ivm.ray2pixel(rxyz, res)
            ppix[:, 0] += res
        else:
            ppix = vm.ray2pixel(rxyz, res)
        rec = np.core.records.fromarrays(ppix.T)
        px, i, cnt = np.unique(rec, return_index=True,
                               return_counts=True)
        cnt = cnt.astype(float)
        omegap = vm.pixel2omega(ppix[i] + .5, res)
        return px, omegap, cnt

    def _smudge(self, cnt, omegap, omegasp):
        """hack to ensure equal energy and max luminance)"""
        ocnt = cnt - (omegap/omegasp)
        smdg = np.sum(ocnt[ocnt > 0])
        cnt[ocnt > 0] = omegap[ocnt > 0]/omegasp
        # average to redistribute
        redist = smdg/np.sum(ocnt < 0)
        # redistribute over pixels with "room" (this could still
        # overshoot if too many pixels are close to threshold, but
        # maybe mathematically impossible?
        cnt[ocnt < -redist] += smdg/np.sum(ocnt < -redist)

    def add_to_img(self, img, pi, sun, vm):
        """
        Parameters
        ----------
        img
        pi
        sun
        vm: raytraverse.mapper.ViewMapper

        Returns
        -------

        """
        res = img.shape[1]
        if pi not in self.vec.keys():
            return None
        # use actual sun location
        # sm = SunMapper(sun[0:3])
        atv = vm.degrees(self.vec[pi])
        if atv <= vm.viewangle/2:
            # rxyz = sm.uv2xyz(r)
            rxyz = self.sunmap.uv2xyz(self.raster[pi], pi[1])
            px, omegap, cnt = self._to_pix(rxyz, atv, vm, res)
            omegasp = self.omega[pi] / self.raster[pi].shape[0]
            self._smudge(cnt, omegap, omegasp)
            # apply average luminanace over each pixel
            clum = sun[-1] * self.lum[pi] * cnt * omegasp / omegap
            for p, cl in zip(px, clum):
                img[tuple(p)] += cl

    def metric(self, psi, v, s, metricfuncs, **kwargs):
        if v.ctheta(s[0:3]) > 0 and psi in self.items():
            svlm = [self.lum[psi]*s[-2]]
            svo = [self.omega[psi]]
            return [f(v, [s[0:3]], svo, svlm, area=2*np.pi,
                    **kwargs) for f in metricfuncs]
        else:
            raise ValueError

    def direct_view(self, res=2):
        """create a summary image of all sun discs from each of vpts"""
        vpts = self.scene.pts()
        ssq = int(np.ceil(np.sqrt(self.suns.suns.shape[0])))
        square = np.array([[0, 0], [0, 1], [1, 1], [1, 0]])
        block = []
        for i in np.arange(self.suns.suns.shape[0], ssq**2):
            sxy = np.unravel_index(i, (ssq, ssq))
            sxy = np.array((sxy[1], sxy[0]))
            block.append(((1, 1, 1), square + sxy))
        for i, pt in enumerate(vpts):
            (lums, fig, ax,
             norm, lev) = plot.mk_img_setup([0, 1], figsize=[ssq*res, ssq*res],
                                            ext=(0, ssq))
            cmap = plot.colormap('viridis', norm)
            plot.plot_patches(ax, block)
            for j in range(self.suns.suns.shape[0]):
                if (i, j) in self.raster.keys():
                    sxy = np.unravel_index(j, (ssq, ssq))
                    sxy = np.array((sxy[1], sxy[0]))
                    lums = cmap.to_rgba(self.lum[(i, j)])
                    xy = (translate.uv2xy(self.raster[(i, j)])+1)/2 + sxy
                    ax.plot(xy[:, 0], xy[:, 1], 'o', ms=2, color=lums)
            ax.set_facecolor((0, 0, 0))
            outf = f"{self.outfile(i)}.png"
            plot.save_img(fig, ax, outf, title=pt)



# -*- coding: utf-8 -*-

# Copyright (c) 2019 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import numpy as np


class MemArrayDict(dict):
    """a dictionary like object that holds arguments for numpy.memmap, the
    getter returns a view to the array"""
    @staticmethod
    def _map(i):
        return np.memmap(i[0], dtype=i[1], mode=i[2], offset=i[3], shape=i[4])

    def __getitem__(self, item):
        return self._map(super().__getitem__(item))

    def values(self):
        return (self._map(item) for item in super().values())

    def constructors(self):
        return (item for item in super().values())

    def full_array(self):
        return self._map(self.full_constructor())

    def full_constructor(self):
        fulli = None
        shape = 0
        for i in self.constructors():
            if fulli is None:
                fulli = list(i)
            shape += i[4][0]
        fulli[4] = (shape, fulli[4][1])
        return tuple(fulli)

    def index_strides(self):
        shape = 0
        strides = [0]
        for i in self.constructors():
            shape += i[4][0]
            strides.append(shape)
        return tuple(strides)



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import os

import numpy as np

from raytraverse import translate, skycalc
from raytraverse.scene.sunsetter import SunSetter


class SunSetterPositions(SunSetter):
    """select suns to sample based on sky pdf and scene.

    Parameters
    ----------
    scene: raytraverse.scene.Scene
        scene class containing geometry, location and analysis plane
    wea: str, np.array, optional
        path to sun position file or wea file, or array of sun positions
    srct: float, optional
        threshold of sky contribution for determining appropriate srcn
    skyro: float, optional
        sky rotation (in degrees, ccw)
    reload: bool
        if True reloads existing sun positions, else always generates new
    """

    def __init__(self, scene, wea, skyro=0.0, **kwargs):
        #: raytraverse.scene.Scene
        self.scene = scene
        #: float: ccw rotation (in degrees) for sky
        self.skyro = skyro
        #: raytraverse.scene.SkyInfo
        self.candidates = wea
        super().__init__(scene, skyro=skyro, **kwargs)

    @property
    def candidates(self):
        """candidate sun positions

        :getter: Returns the sun source array
        :setter: Set the sun source array and write to files
        :type: np.array
        """
        return self._candidates

    @candidates.setter
    def candidates(self, wea):
        """load candidate sun positions"""
        try:
            dat = np.loadtxt(wea)
            if dat.shape[1] == 3:
                dummydir = np.ones((dat.shape[0], 1))
                self._candidates = np.hstack((translate.norm(dat), dummydir))
            elif dat.shape[1] == 4:
                xyz = translate.aa2xyz(dat[:, 0:2])
                self._candidates = np.hstack((xyz, dat[:, 3:4]))
            elif dat.shape[1] == 5 and np.max(dat[:, 0:3] <= 1):
                xyz = translate.norm(dat[:, 0:3])
                self._candidates = np.hstack((xyz, dat[:, 3:4]))
            else:
                raise ValueError
        except ValueError:
            loc = skycalc.get_loc_epw(wea)
            wdat = skycalc.read_epw(wea)
            times = skycalc.row_2_datetime64(wdat[:, 0:3])
            xyz = skycalc.sunpos_xyz(times, *loc, ro=self.skyro)
            self._candidates = np.hstack((xyz, wdat[:, 3:4]))

    def choose_suns(self):
        uvsize = self.sunres
        cbins = translate.uv2bin(translate.xyz2uv(self.candidates[:, 0:3],
                                                  normalize=True, flipu=False),
                                 self.scene.skyres)
        cidxs = np.arange(cbins.size)
        sbins = np.arange(uvsize**2)
        skyb = self.load_sky_facs()
        if skyb is 1:
            skyb = np.full(sbins.shape[0], 1)
        idxs = []
        for b in sbins:
            if skyb[b] > self.srct:
                try:
                    # choose sun position, priortizing brighter conditions
                    # may introduce bias?
                    p = self.candidates[cbins == b, 3]
                    sp = np.sum(p)
                    if sp == 0:
                        raise ValueError
                    a = np.random.choice(cidxs[cbins == b], p=p/sp)
                except ValueError:
                    pass
                else:
                    idxs.append(a)
        return self.candidates[idxs, 0:3]




# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import sys

import numpy as np

from raytraverse import translate
from raytraverse.sampler import SingleSunSampler, SunViewSampler


class SunSampler(object):
    """factory class for SingleSunSamplers.

    Parameters
    ----------
    scene: raytraverse.scene.Scene
        scene class containing geometry, location and analysis plane
    suns: raytraverse.sunsetter.SunSetter
        sun class containing sun locations.
    """

    def __init__(self, scene, suns, plotp=False, **kwargs):
        self.scene = scene
        self.suns = suns
        #: raytraverse.sampler.SunViewSampler
        self.viewsampler = SunViewSampler(scene, suns, plotp=False)
        #: dict: sampling arguments for SingleSunSampler
        self.sampleargs = dict(idres=4, fdres=10, speclevel=9, plotp=plotp)
        #: raytraverse.sampler.SingleSunSampler
        self.reflsampler = None
        self.sampleargs.update(**kwargs)

    def run(self, view=True, reflection=True):
        if view and self.suns.suns.size > 0:
            print("Sampling Sun Visibility", file=sys.stderr)
            self.viewsampler.run()
        if reflection:
            suncnt = self.suns.suns.shape[0]
            aa = translate.xyz2aa(self.suns.suns)
            for sidx in range(suncnt):
                print(f"Sampling Sun Reflections {sidx+1} of "
                      f"{suncnt}", file=sys.stderr)
                print(f'Sun Position: alt={aa[sidx, 0]:.01f},'
                      f' az={aa[sidx, 1]:.01f}', file=sys.stderr)
                self.reflsampler = SingleSunSampler(self.scene, self.suns, sidx,
                                                    **self.sampleargs)
                self.reflsampler.run()



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

"""renderer objects"""

__all__ = ['Renderer', 'SPRenderer', 'RadianceRenderer', 'Rtrace', 'Rcontrib',
           'SPRtrace', 'SPRcontrib']

from raytraverse.renderer.renderer import Renderer
from raytraverse.renderer.sprenderer import SPRenderer, SPRcontrib, SPRtrace
from raytraverse.renderer.radiancerenderer import RadianceRenderer
from raytraverse.renderer.rtrace import Rtrace
from raytraverse.renderer.rcontrib import Rcontrib



# -*- coding: utf-8 -*-

# Copyright (c) 2019 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import numpy as np

from raytraverse import translate


class ViewMapper(object):
    """translate between world and normalized UV space based on direction
    and view angle

    Parameters
    ----------
    dxyz: tuple, optional
        central view direction
    viewangle: float, optional
        if < 180, the horizontal and vertical view angle, if greater, view
        becomes 360,180
    """

    def __init__(self, dxyz=(0.0, 1.0, 0.0), viewangle=360.0, name='view',
                 mtxs=None, imtxs=None):
        self._viewangle = viewangle
        # float: aspect ratio width/height
        self.aspect = 1
        self._initmtx = (mtxs, imtxs)
        self.dxyz = dxyz
        self._initmtx = None
        self.area = 2*np.pi*(1 - np.cos(viewangle*np.pi/360))
        self.name = name

    def __getitem__(self, item):
        mtxs = [(self.ymtx[item], self.pmtx[item])]
        if self._ivm is not None:
            imtxs = [(self.ivm.ymtx[item], self.ivm.pmtx[item])]
        else:
            imtxs = None
        return ViewMapper(self.dxyz[item], self.viewangle, self.name,
                          mtxs=mtxs, imtxs=imtxs)

    @property
    def dxyz(self):
        """(float, float, float) central view direction"""
        return self._dxyz

    @property
    def viewangle(self):
        """view angle"""
        return self._viewangle

    @property
    def ymtx(self):
        """yaw rotation matrix (to standard z-direction y-up)"""
        return self._ymtx

    @property
    def pmtx(self):
        """pitch rotation matrix (to standard z-direction y-up)"""
        return self._pmtx

    @property
    def bbox(self):
        """np.array of shape (2,2): bounding box of view"""
        return self._bbox

    @property
    def sf(self):
        """bbox scale factor"""
        return self._sf

    @property
    def ivm(self):
        """viewmapper for opposite view direction (in case of 360 degree view"""
        return self._ivm

    @dxyz.setter
    def dxyz(self, xyz):
        """set view parameters"""
        self._dxyz = translate.norm(xyz)
        if self._initmtx[0] is None:
            self._ymtx, self._pmtx = zip(*[translate.rmtx_yp(x)
                                           for x in self.dxyz])
        else:
            self._ymtx, self._pmtx = zip(*self._initmtx[0])
        if self._viewangle > 180:
            self._sf = np.array((1, 1))
            self._bbox = np.stack(((0, 0), (2, 1)))
            self.aspect = 2
            self._ivm = ViewMapper(-self.dxyz, 180, mtxs=self._initmtx[1])
        else:
            self._sf = np.array((self._viewangle/180, self._viewangle/180))
            self._bbox = np.stack((.5 - self._sf/2, .5 + self._sf/2))
            self._ivm = None

    def view2world(self, xyz, i=0):
        return (self.ymtx[i].T@(self.pmtx[i].T@xyz.T)).T

    def world2view(self, xyz, i=0):
        return (self.pmtx[i]@(self.ymtx[i]@xyz.T)).T

    def xyz2uv(self, xyz, i=0):
        # rotate from world to view space
        rxyz = self.world2view(xyz, i)
        # translate to uv
        uv = translate.xyz2uv(rxyz)
        # scale to view uv space
        uv = (uv - self.bbox[None, 0])/self.sf[None, :]
        return uv

    def uv2xyz(self, uv, i=0):
        # scale to hemispheric uv
        uv = self.bbox[None, 0] + uv*self.sf[None, :]
        # translate to xyz with z view direction
        rxyz = translate.uv2xyz(uv)
        # rotate from this view space to world
        if np.asarray(i).size > 1:
            xyz = np.concatenate([self.view2world(r.reshape(1, -1), j)
                                  for r, j in zip(rxyz, i)])
        else:
            xyz = self.view2world(rxyz, i)
        return xyz

    def xyz2xy(self, xyz, i=0):
        rxyz = self.world2view(xyz, i)
        return translate.xyz2xy(rxyz)

    def pixelrays(self, res, i=0):
        pxy = np.stack(np.mgrid[0:res, 0:res], 2) + .5
        if self.aspect == 2:
            a = self.pixel2ray(pxy, res, i)
            b = self.ivm.pixel2ray(pxy, res, i)
            return np.concatenate((a, b), 0)
        else:
            return self.pixel2ray(pxy, res, i)

    def ray2pixel(self, xyz, res, i=0):
        xy = self.xyz2xy(xyz, i)
        pxy = np.floor((xy/2 + .5) * res).astype(int)
        pxy[:, 0] = res - 1 - pxy[:, 0]
        return pxy

    def pixel2ray(self, pxy, res, i=0):
        rxyz = translate.pxy2xyz(pxy/res, self.viewangle/self.aspect)
        xyz = self.view2world(rxyz.reshape(-1, 3), i).reshape(rxyz.shape)
        return xyz

    def pixel2omega(self, pxy, res):
        va = self.viewangle/self.aspect
        of = np.array(((.5, 0), (0, .5)))
        xa = translate.pxy2xyz((pxy - of[0])/res, va)
        xb = translate.pxy2xyz((pxy + of[0])/res, va)
        ya = translate.pxy2xyz((pxy - of[1])/res, va)
        yb = translate.pxy2xyz((pxy + of[1])/res, va)
        cp = np.cross(xb - xa, yb - ya)
        return np.linalg.norm(cp, axis=-1)

    def ctheta(self, vec, i=0):
        vec = np.asarray(vec)
        return np.einsum("i,ji->j", self.dxyz[i],
                         vec.reshape(-1, vec.shape[-1]))

    def radians(self, vec, i=0):
        return np.arccos(self.ctheta(vec, i))

    def degrees(self, vec, i=0):
        return self.radians(vec, i) * 180/np.pi

    def in_view(self, vec, i=0, indices=True):
        ang = self.radians(vec, i)
        mask = ang < self.viewangle*np.pi/360/self.aspect
        if indices:
            return np.unravel_index(np.arange(ang.size)[mask], vec.shape[:-1])
        else:
            return mask



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import _pytest.skipping


def pytest_addoption(parser):
    parser.addoption(
        "--slowtest",
        action="store_true",
        default=False, help="disable skip marks")


@pytest.hookimpl(tryfirst=True)
def pytest_cmdline_preparse(config, args):
    if "--slowtest" not in args:
        return

    def no_skip(*args, **kwargs):
        return

    _pytest.skipping.skip = no_skip



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

"""c++ renderer objects"""

__all__ = ['cRcontrib', 'cRtrace']

from raytraverse.crenderer.rcontrib_c import cRcontrib
from raytraverse.crenderer.rtrace_c import cRtrace



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

import os
import pickle
import sys
from concurrent.futures import ThreadPoolExecutor

import numpy as np
from scipy.spatial import cKDTree

from raytraverse import translate, plot
from raytraverse.mapper import SunMapper


class SunSetter(object):
    """select suns to sample based on sky pdf and scene.

    Parameters
    ----------
    scene: raytraverse.scene.Scene
        scene class containing geometry, location and analysis plane
    srct: float, optional
        threshold of sky contribution for determining appropriate srcn
    skyro: float, optional
        sky rotation (in degrees, ccw)
    reload: bool
        if True reloads existing sun positions, else always generates new
    """

    def __init__(self, scene, srct=.01, skyro=0.0, reload=True, sunres=10.0,
                 **kwargs):
        #: float: threshold of sky contribution for determining appropriate srcn
        self.srct = srct
        #: float: ccw rotation (in degrees) for sky
        self.skyro = skyro
        self.sunres = sunres
        sunfile = f"{scene.outdir}/suns.rad"
        #: bool: reuse existing sun positions (if found)
        if not reload:
            try:
                os.remove(sunfile)
            except FileNotFoundError:
                pass
            try:
                os.remove(f'{scene.outdir}/sky_skydetail.npy')
            except FileNotFoundError:
                pass
        #: raytraverse.scene.Scene
        self.scene = scene
        self.suns = sunfile
        if self.suns.size == 0:
            print('Warning! no suns set, check parameters and sky detail!',
                  file=sys.stderr)
            self.map = None
        else:
            self.map = SunMapper(self.suns)
        self.sun_kd = None

    @property
    def sunres(self):
        return self._sunres

    @sunres.setter
    def sunres(self, s):
        self._sunres = int(np.floor(90/s)*2)

    @property
    def sun_kd(self):
        """sun kdtree for directional queries"""
        if self._sun_kd is None:
            self._sun_kd = cKDTree(self.suns)
        return self._sun_kd

    @sun_kd.setter
    def sun_kd(self, sun_kd):
        self._sun_kd = sun_kd

    @property
    def suns(self):
        """holds sun positions

        :getter: Returns the sun source array
        :setter: Set the sun source array and write to files
        :type: np.array
        """
        return self._suns

    @suns.setter
    def suns(self, sunfile):
        """set the skydetail array and determine sample count and spacing"""
        if os.path.isfile(sunfile):
            f = open(sunfile, 'r')
            sund = f.read().split('source')[1:]
            xyz = np.array([s.split()[4:7] for s in sund]).astype(float)
            self._suns = xyz
        else:
            self._suns = self.choose_suns()
            self._write_suns(sunfile)

    def choose_suns(self):
        uvsize = self.sunres
        si = np.stack(np.unravel_index(np.arange(uvsize**2),
                                       (uvsize, uvsize)))
        skyb = self.load_sky_facs()
        print(uvsize)
        uv = si.T/uvsize
        ib = np.full(uv.shape[0], True)
        ib = (skyb*ib > self.srct)
        border = 2*uvsize/180
        j = np.random.default_rng().uniform(border, 1 - border, si.shape)
        # j = .5
        uv = (si + j).T[ib]/uvsize
        xyz = translate.uv2xyz(uv, xsign=1)
        return xyz

    def load_sky_facs(self):
        outf = f'{self.scene.outdir}/sky_skydetail.npy'
        if os.path.isfile(outf):
            sd = np.load(outf)
        else:
            dfile = f'{self.scene.outdir}/sky_kd_lum_map.pickle'
            try:
                f = open(dfile, 'rb')
            except FileNotFoundError:
                print('Warning! suns initialized without sky detail, first'
                      ' create a SCBinField', file=sys.stderr)
                return 1
            else:
                skylums = pickle.load(f)
                f.close()
                zeros = np.zeros(len(skylums), dtype=int)
                with ThreadPoolExecutor() as exc:
                    mxs = list(exc.map(np.max, skylums.values(), zeros))
                sd = np.max(np.stack(mxs), 0).reshape(self.scene.skyres,
                                                      self.scene.skyres)
                np.save(outf, sd)
        sd = translate.interpolate2d(sd, (self.sunres, self.sunres))
        return sd.ravel()

    def direct_view(self):
        sxy = translate.xyz2xy(self.suns, flip=False)
        sbins = translate.uv2bin(translate.xyz2uv(self.suns, flipu=False),
                                 self.sunres)
        lums, fig, ax, norm, lev = plot.mk_img_setup([0, 1], ext=1)
        outf = f'{self.scene.outdir}/sky_skydetail.npy'
        if os.path.isfile(outf):
            sf = translate.interpolate2d(np.load(outf),
                                         (self.sunres, self.sunres)).ravel()
            borders = translate.bin_borders(np.arange(sf.size),
                                            self.sunres)
            sfxyz = translate.uv2xyz(borders.reshape(-1, 2), xsign=1)
            sfxy = translate.xyz2xy(sfxyz, flip=False).reshape(-1, 4, 2)
            cmap = plot.colormap('gray',
                                 plot.Normalize(vmin=np.log10(self.srct) - 1,
                                                vmax=0))
            colors = cmap.to_rgba(np.log10(sf))
            patcha = [(c, p) for c, p in zip(colors, sfxy)]
            patchb = [(c, p) for c, p in zip(colors[sbins], sfxy[sbins])]
            plot.plot_patches(ax, patcha, {'zorder': -2, 'lw': .25,
                                           'edgecolor': (.2, .2, .2)})
            plot.plot_patches(ax, patchb, {'zorder': -1, 'lw': 3,
                                           'edgecolor': (1, .5, 0)})
        ax.plot(sxy[:, 0], sxy[:, 1], lw=.5, ms=0, color='grey')
        ax.scatter(sxy[:, 0], sxy[:, 1], s=30, marker='o', color=(1, .5, 0))
        ax.set_facecolor((.2, .3, .5))
        outf = f"{self.scene.outdir}_suns.png"
        plot.save_img(fig, ax, outf)

    def write_sun(self, i):
        s = self.suns[i]
        mod = f"solar{i:05d}"
        name = f"sun{i:05d}"
        d = f"{s[0]} {s[1]} {s[2]}"
        dec = f"void light {mod} 0 0 3 1 1 1\n"
        dec += f"{mod} source {name} 0 0 4 {d} 0.533\n"
        return dec

    def proxy_src(self, tsuns, tol=10.0):
        """check if sun directions have matching source in SunSetter

        Parameters
        ----------
        tsuns: np.array
            (N, 3) array containing sun source vectors to check
        tol: float
            tolerance (in degrees)

        Returns
        -------
        np.array
            (N,) boolean array if sun has a match
        np.array
            (N,) index to proxy src
        """
        stol = translate.theta2chord(tol*np.pi/180)
        suns = translate.norm(tsuns)
        serrs, sis = self.sun_kd.query(suns)
        sis = np.where(serrs < stol, sis, self.suns.shape[0])
        return sis

    def _write_suns(self, sunfile):
        """write suns to file

        Parameters
        ----------
        sunfile
        """
        if self.suns.size > 0:
            f = open(sunfile, 'w')
            for i in range(self.suns.shape[0]):
                dec = self.write_sun(i)
                print(dec, file=f)
            f.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================


"""entry point for genskyvec."""
import subprocess
import sys


def main():
    sys.argv[0] = __file__.rsplit('/', 1)[0] + '/genskyvec.pl'
    subprocess.call(sys.argv)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for raytraverse.scene"""
import os
import shutil

import pytest
from raytraverse.translate import scbinscal, scxyzcal
from raytraverse.scene import Scene
from raytraverse.sampler import SCBinSampler
from raytraverse import translate
import numpy as np
import clasp.script_tools as cst


# from clipt import mplt


@pytest.fixture(scope="module")
def tmpdir(tmp_path_factory):
    data = str(tmp_path_factory.mktemp("data"))
    shutil.copytree('tests/test/', data + '/test')
    shutil.copy('raytraverse/genskyvec.pl', data + '/test')
    cpath = os.getcwd()
    os.chdir(data + '/test')
    yield data + '/test'
    os.chdir(cpath)


@pytest.fixture
def scene(tmpdir):
    return Scene('results', 'test.oct', 'plane.rad', wea='geneva.epw')


def test_cals(tmpdir):
    f = open('d2b.cal', 'w')
    f.write(scbinscal)
    f.close()
    f = open('b2d.cal', 'w')
    f.write(scxyzcal)
    f.close()
    bins = np.arange(100)
    result = cst.pipeline(
        ["cnt 100 | rcalc -f rayinit.cal -f b2d.cal -e 'side=10;bin=$1"
         ";$1=$1;$2=U;$3=V;$4=Dx;$5=Dy;$6=Dz' | rcalc -f rayinit.cal -f "
         "d2b.cal -e 'side=10;Dx=$4;Dy=$5;Dz=$6;$1=bin'"])
    r = [float(i) for i in result.split()]
    assert np.allclose(r, bins)


def test_init(tmpdir, scene):
    res = np.array([[32, 16, ],
                    [64, 32, ],
                    [128, 64, ],
                    [256, 128, ],
                    [512, 256, ],
                    [1024, 512, ]])
    sampler = SCBinSampler(scene)
    f = open(f'{sampler.scene.outdir}/scbins.cal')
    assert f.read() == scbinscal
    f.close()
    assert np.alltrue(res == sampler.levels)


def test_sky_sample(tmpdir, scene, capfd):
    sampler = SCBinSampler(scene)
    vecf = sampler.dump_vecs(np.array([[5, 5, 1.25, 0, -1, 0]]))
    with capfd.disabled():
        lum = sampler.sample(vecf)
    assert lum > .05/179



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for raytraverse.scene"""
import os
import shutil

import pytest

from raytraverse.mapper import SpaceMapperPt
from raytraverse.scene import Scene, SkyInfo
from raytraverse import quickplot
import numpy as np
from clipt import mplt




@pytest.fixture(scope="module")
def tmpdir(tmp_path_factory):
    data = str(tmp_path_factory.mktemp("data"))
    shutil.copytree('tests/test/', data + '/test')
    cpath = os.getcwd()
    os.chdir(data + '/test')
    yield data + '/test'
    os.chdir(cpath)


def test_scene(tmpdir):
    scene = Scene('results', 'test.oct', 'plane.rad')
    assert scene.scene == 'results/scene.oct'
    scene2 = Scene('results', 'MATERIAL/*.mat RAD/*.rad', 'plane.rad', overwrite=True, reload=False)
    assert scene2.scene == 'results/scene.oct'
    with pytest.raises(ChildProcessError):
        s = Scene('results', 'RAD/*.rad', 'plane.rad', overwrite=True)
        os.system(f'getinfo {s.scene}')
    with pytest.raises(FileExistsError):
        Scene('results', 'test.oct', 'plane.rad', reload=False)


def test_solarbounds(tmpdir):
    loc = (46.25, -6.13, -15)
    scene = SkyInfo(loc)
    assert np.all(np.logical_not(scene.in_solarbounds(np.array([[.5,.5], [1.5,1], [.5,0]]))))


def test_area(tmpdir):
    scene = Scene('results', 'test.oct', 'plane.rad', overwrite=True)
    # print(scene.area.bbox)
    # print(np.prod(scene.area.bbox[1, 0:2] - scene.area.bbox[0, 0:2]))
    assert scene.area.in_area(np.array([[2.5, 3.5, 1.25]]))
    assert not scene.area.in_area(np.array([[0, 0, 0]]))
    scene = Scene('results', 'test.oct', 'plane.rad', overwrite=True, ptro=3)
    assert scene.area.in_area(np.array([[2.5, 3.5, 1.25]]))
    grid_u, grid_v = np.meshgrid(np.arange(.0005, 1, .001), np.arange(.0005, 1, .001))
    uv = np.vstack((grid_u.flatten(), grid_v.flatten())).T
    xyz = scene.area.uv2pt(uv)
    ia = scene.area.in_area(xyz)
    # quickplot.imshow(ia.reshape(1000, 1000))
    area = np.prod(scene.area.bbox[1, 0:2] - scene.area.bbox[0, 0:2])*np.sum(ia)/uv.shape[0]
    assert np.isclose(area, 171, atol=.05)
    assert not scene.area.in_area(np.array([[0, 19, 0]]))


def test_areapt(tmpdir):
    np.savetxt('pts.txt', np.random.random((10, 3))*100)
    scene = Scene('results', 'test.oct', 'pts.txt', overwrite=True)
    sp = scene.area
    pt0 = sp.pts()
    uv0 = sp.pt2uv(pt0)
    b = sp.uv2pt(sp.pt2uv(pt0))
    uve = np.arange(0.05, 1, .1)
    assert np.allclose(pt0, b)
    assert np.allclose(uv0[:, 0], uve)



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import numpy as np

from raytraverse.lightfield.lightfieldkd import LightFieldKD


class SCBinField(LightFieldKD):
    """container for accessing sampled data where every ray has a value for
    each source
    """

    def __init__(self, scene, rebuild=False, prefix='sky', **kwargs):
        super().__init__(scene, rebuild=rebuild, prefix=prefix,
                         srcn=scene.skyres**2, **kwargs)

    def apply_coef(self, pi, coefs):
        coefs = np.asarray(coefs)
        if np.mod(coefs.size, self.srcn) == 0:
            c = coefs.reshape(-1, self.srcn)
        else:
            c = np.broadcast_to(coefs, (coefs.size, self.srcn))
        return np.einsum('ij,kj->ik', c, self.lum[pi])




# -*- coding: utf-8 -*-

# Copyright (c) 2019 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

"""functions for translating between coordinate spaces and resolutions"""

import numpy as np
from scipy.interpolate import RectBivariateSpline
from scipy.ndimage.filters import gaussian_filter, uniform_filter

scbinscal = ("""
{ map U/V axis to bin divisions }
axis(x) : mod(floor(side * x), side);

{ get bin of u,v }
binl(u, v) : axis(u)*side + axis(v);

{ shirley-chiu disk to square (with spherical term) }
pi4 : PI/4;
n = if(Dz, 1, -1);
r2 = 1 - n*Dz;
x = Dx/sqrt(2 - r2);
y = -Dy/sqrt(2 - r2);
r = sqrt( sq(x) + sq(y));
ph = atan2(x, y);
phi = ph + if(-pi4 - ph, 2*PI, 0);
a = if(pi4 - phi, r, if(3*pi4 - phi, -(phi - PI/2)*r/pi4, if(5*pi4 - phi,"""
             """ -r, (phi - 3*PI/2)*r/pi4)));
b = if(pi4 - phi, phi*r/pi4, if(3*pi4 - phi, r, if(5*pi4 - phi, """
             """-(phi - PI)*r/pi4, -r)));

{ map to (0,2),(0,1) matches raytraverse.translate.xyz2uv}
U = (if(n, 1, 3) - a*n)/2;
V = (b + 1)/2;

bin = binl(V, U);
""")

scxyzcal = """
x1 = .5;
x2 = .5;

U = ((bin - mod(bin, side)) / side + x1)/side;
V = (mod(bin, side) + x2)/side;

n = if(U - 1, -1, 1);
ur = if(U - 1, U - 1, U);
a = 2 * ur - 1;
b = 2 * V - 1;
conda = sq(a) - sq(b);
condb = abs(b) - FTINY;
r = if(conda, a, if(condb, b, 0));
phi = if(conda, b/(2*a), if(condb, 1 - a/(2*b), 0)) * PI/2;
sphterm = r * sqrt(2 - sq(r));
Dx = n * cos(phi)*sphterm;
Dy = sin(phi)*sphterm;
Dz = n * (1 - sq(r));
"""


def norm(v):
    """normalize 2D array of vectors along last dimension"""
    return v/np.linalg.norm(v, axis=-1).reshape(-1, 1)


def norm1(v):
    """normalize flat vector"""
    return v/np.sqrt(np.sum(np.square(v)))


def tpnorm(thetaphi):
    """normalize angular vector to 0-pi, 0-2pi"""
    thetaphi[:, 0] = np.mod(thetaphi[:, 0] + np.pi, np.pi)
    thetaphi[:, 1] = np.mod(thetaphi[:, 1] + 2*np.pi, 2*np.pi)
    return thetaphi


def uv2xy(uv):
    """translate from unit square (0,1),(0,1) to disk (x,y)
    http://psgraphics.blogspot.com/2011/01/improved-code-for-concentric
    -map.html.
    """
    np.seterr(all="ignore")
    xy = np.empty((uv.shape[0], 2), np.float)
    u = uv[:, 0]
    v = uv[:, 1]
    n = np.where(u > 1.0, -1, 1)
    u = np.where(u > 1.0, u - 1.0, u)
    a = 2. * u - 1
    b = 2. * v - 1
    cond = a*a > b*b
    r = np.where(cond, a, np.where(b == 0, 0, b))
    phi = np.where(cond, b/(2*a), np.where(b == 0, 0, 1 - a/(2*b)))*np.pi/2
    xy[:, 0] = n*np.cos(phi)*r
    xy[:, 1] = np.sin(phi)*r
    return xy


def uv2xyz(uv, axes=(0, 1, 2), xsign=-1):
    """translate from 2 x unit square (0,2),(0,1) to unit sphere (x,y,z)
    http://psgraphics.blogspot.com/2011/01/improved-code-for-concentric
    -map.html.
    """
    np.seterr(all="ignore")
    xyz = np.empty((uv.shape[0], 3), np.float)
    u = uv[:, 0]
    v = uv[:, 1]
    # u > 1 values lay in the negative z hemisphere
    n = np.where(u > 1.0, -1, 1)
    # bring both hemispheres to (0,1)
    u = np.where(u > 1.0, u - 1.0, u)
    a = 2. * u - 1
    b = 2. * v - 1
    cond = a*a > b*b
    r = np.where(cond, a, np.where(b == 0, 0, b))
    phi = np.where(cond, b/(2*a), np.where(b == 0, 0, 1 - a/(2*b)))*np.pi/2
    sphterm = r*np.sqrt(2 - r*r)
    # flip back x in positive z space
    xyz[:, axes[0]] = xsign*n*np.cos(phi)*sphterm
    xyz[:, axes[1]] = np.sin(phi)*sphterm
    # add sign to z
    xyz[:, axes[2]] = n*(1 - r*r)
    return xyz


def xyz2uv(xyz, normalize=False, axes=(0, 1, 2), flipu=True):
    """translate from vector x,y,z (normalized) to u,v (0,2),(0,1)
    Shirley, Peter, and Kenneth Chiu. A Low Distortion Map Between Disk and
    Square. Journal of Graphics Tools, vol. 2, no. 3, Jan. 1997, pp. 45-52.
    Taylor and Francis+NEJM, doi:10.1080/10867651.1997.10487479.
    """
    if normalize:
        xyz = norm(xyz)
    uv = np.empty((xyz.shape[0], 2), np.float)
    # store sign of z-axis to map both hemispheres as positive
    n = np.where(xyz[:, axes[2]] < 0, -1, 1)
    r2 = 1 - n*xyz[:, axes[2]]
    x = xyz[:, axes[0]] / np.sqrt(2 - r2)
    y = xyz[:, axes[1]] / np.sqrt(2 - r2)
    r = np.sqrt(np.square(x) + np.square(y))
    phi = np.arctan2(y, x)
    pi4 = np.pi/4
    phi = phi + np.where(phi < -pi4, 2*np.pi, 0)
    a = np.where(phi < pi4, (r, phi*r/pi4),
                 np.where(phi < 3*pi4, (-(phi - np.pi/2)*r/pi4, r),
                          np.where(phi < 5*pi4, (-r, -(phi - np.pi)*r/pi4),
                                   ((phi - 3*np.pi/2)*r/pi4, -r)))).T
    # for the positive z-direction (n=1) map -1,1 to 1,0 (to correct flip)
    # for the negative z-direction (n=-1) map -1,1 to 1,2
    if flipu:
        uv[:, 0] = (np.where(n < 0, 3, 1) - a[:, 0]*n) / 2.
    else:
        uv[:, 0] = (a[:, 0] + 2 - n) / 2.
    uv[:, 1] = (a[:, 1] + 1) / 2.
    return uv


def xyz2xy(xyz, axes=(0, 1, 2), flip=True):
    r = np.arctan2(np.sqrt(np.sum(np.square(xyz[:, axes[0:2]]), -1)),
                   xyz[:, axes[2]])/(np.pi/2)
    phi = np.arctan2(xyz[:, axes[0]], xyz[:, axes[1]])
    x = r * np.sin(phi)
    y = r * np.cos(phi)
    if flip:
        x = -x
    return np.stack((x, y)).T


def pxy2xyz(pxy, viewangle=180.0):
    pxy -= .5
    pxy *= viewangle/180
    d = np.sqrt(np.sum(np.square(pxy), -1))
    z = np.cos(np.pi*d)
    d = np.where(d <= 0, np.pi, np.sqrt(1 - z*z)/d)
    pxy *= d[..., None]
    xyz = np.concatenate((pxy, z[..., None]), -1)
    return xyz


def tp2xyz(thetaphi, normalize=True):
    """calculate x,y,z vector from theta (0-pi) and phi (0-2pi) RHS Z-up"""
    if normalize:
        thetaphi = tpnorm(thetaphi)
    theta = thetaphi[:, 0]
    phi = thetaphi[:, 1]
    sint = np.sin(theta)
    xyz = np.array([sint*np.cos(phi), sint*np.sin(phi),
                    np.cos(theta)]).T
    return norm(xyz)


def xyz2tp(xyz):
    """calculate theta (0-pi), phi from x,y,z RHS Z-up"""
    theta = np.arccos(xyz[:, 2])
    phi = np.where(np.isclose(theta, 0.0, atol=1e-10), np.pi,
                   np.where(np.isclose(theta, np.pi, atol=1e-10),
                            np.pi, np.arctan2(xyz[:, 1], xyz[:, 0])))
    return tpnorm(np.column_stack([theta, phi]))


def tp2uv(thetaphi):
    """calculate UV from theta (0-pi), phi"""
    return xyz2uv(tp2xyz(thetaphi))


def uv2tp(uv):
    """calculate theta (0-pi), phi from UV"""
    return xyz2tp(uv2xyz(uv))


def uv2ij(uv, side):
    ij = np.mod(np.floor(side*uv), side)
    ij[:, 0] += (uv[:, 0] >= 1) * side
    return ij.astype(int)


def uv2bin(uv, side):
    buv = uv2ij(uv, side)
    return buv[:, 0]*side + buv[:, 1]


def bin2uv(bn, side):
    u = (bn - np.mod(bn, side))/(side*side)
    v = np.mod(bn, side)/side
    return np.stack((u, v)).T


def bin_borders(sb, side):
    si = np.stack(np.unravel_index(sb, (side, side)))
    square = np.array([[0, 0], [0, 1], [1, 1], [1, 0]])/side
    uv = np.repeat(si.T[:, None, :]/side, 4, axis=1) + square[None]
    return uv


def resample(samps, ts=None, gauss=True, radius=None):
    """simple array resampling. requires whole number multiple scaling.

    Parameters
    ----------
    samps: np.array
        array to resample along each axis
    ts: tuple, optional
        shape of output array, should be multiple of samps.shape
    gauss: bool, optional
        apply gaussian filter to upsampling
    radius: float, optional
        when gauss is True, filter radius, default is the scale ratio - 1

    Returns
    -------
    np.array
        to resampled array

    """
    if ts is None:
        ts = samps.shape
    rs = np.array(ts)/np.array(samps.shape)
    if np.prod(rs) > 1:
        for i in range(len(rs)):
            samps = np.repeat(samps, rs[i], i)
        if gauss:
            if radius is None:
                radius = tuple(rs - 1)
            samps = gaussian_filter(samps, radius)
    elif np.prod(rs) < 1:
        rs = (1/rs).astype(int)
        og = (-rs/2).astype(int)
        samps = uniform_filter(samps, rs, origin=og)
        for i, j in enumerate(rs):
            samps = np.take(samps, np.arange(0, samps.shape[i], j), i)
    elif radius is not None:
        if gauss:
            samps = gaussian_filter(samps, radius)
        else:
            samps = uniform_filter(samps, int(radius))
    return samps


def interpolate2d(a, s):
    oldcx = np.linspace(0, 1, a.shape[0])
    newcx = np.linspace(0, 1, s[0])
    oldcy = np.linspace(0, 1, a.shape[1])
    newcy = np.linspace(0, 1, s[1])
    f = RectBivariateSpline(oldcx, oldcy, a, kx=1, ky=1)
    return f(newcx, newcy)


def rmtx_elem(theta, axis=2, degrees=True):
    if degrees:
        theta = theta * np.pi / 180
    rmtx = np.array([(np.cos(theta), -np.sin(theta), 0),
                     (np.sin(theta), np.cos(theta), 0),
                     (0, 0, 1)])
    return np.roll(rmtx, axis-2, (0, 1))


def rotate_elem(v, theta, axis=2, degrees=True):
    rmtx = rmtx_elem(theta, axis=axis, degrees=degrees)
    return np.einsum('ij,kj->ki', rmtx, v)


def rmtx_yp(v):
    """generate a pair of rotation matrices to transform from vector v to
    z, enforcing a z-up in the source space and a y-up in the destination. If
    v is z, returns pair of identity matrices, if v is -z returns pair of 180
    degree rotation matrices.

    Parameters
    ----------
    v: array-like of size (3,)
        the vector direction representing the starting coordinate space

    Returns
    -------

    ymtx, pmtx: (np.array, np.array)
        two rotation matrices to be premultiplied in order to reverse transorm,
        swap order and transpose.
        Forward: pmtx@(ymtx@xyz.T)).T
        Backward: ymtx.T@(pmtx.T@xyz.T)).T
    """
    v = norm1(v)
    v2 = np.array((0, 0, 1))
    if np.isnan(v[0]):
        raise ValueError(f"Vector Normalization Failed: {v}")
    if np.allclose(v, v2):
        return np.identity(3), np.identity(3)
    elif np.allclose(v, -v2):
        ymtx = np.array([(-1, 0, 0),
                         (0, -1, 0),
                         (0, 0, 1)])
        pmtx = np.array([(1, 0, 0),
                         (0, -1, 0),
                         (0, 0, -1)])
        return ymtx, pmtx
    tp = xyz2tp(v.reshape(-1, 3))[0]
    y = 3*np.pi/2 - tp[1]
    p = -tp[0]
    ymtx = np.array([(np.cos(y), -np.sin(y), 0),
                     (np.sin(y), np.cos(y), 0),
                     (0, 0, 1)])
    pmtx = np.array([(1, 0, 0),
                     (0, np.cos(p), -np.sin(p)),
                     (0, np.sin(p), np.cos(p))])
    return ymtx, pmtx


def chord2theta(c):
    """compute angle from chord on unit circle

    Parameters
    ----------
    c: float
        chord or euclidean distance between normalized direction vectors

    Returns
    -------
    theta: float
        angle captured by chord
    """
    return 2*np.arcsin(c/2)


def theta2chord(theta):
    """compute chord length on unit sphere from angle

    Parameters
    ----------
    theta: float
        angle

    Returns
    -------
    c: float
        chord or euclidean distance between normalized direction vectors
    """
    return 2*np.sin(theta/2)


def aa2xyz(aa):
    tp = np.pi/2 - aa * np.pi/180
    tp[:, 1] += np.pi
    return tp2xyz(tp)


def xyz2aa(xyz):
    tp = xyz2tp(xyz)
    tp[:, 1] -= np.pi
    return (np.pi/2 - tp)/(np.pi/180)



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================


"""Console script for raytraverse."""
import os
import sys

import numpy as np

from clasp import click
import clasp.click_ext as clk
import raytraverse
from raytraverse.integrator import Integrator
from raytraverse.sampler import SCBinSampler, SunSampler
from raytraverse.scene import Scene, SunSetter, SunSetterLoc, SunSetterPositions
from raytraverse.lightfield import SCBinField, SunField, SunViewField


@clk.pretty_name("NPY, TSV, FLOATS,FLOATS")
def np_load(ctx, param, s):
    """read np array from command line

    trys np.load (numpy binary), then np.loadtxt (space seperated txt file)
    then split row by spaces and columns by commas.
    """
    if os.path.exists(s):
        try:
            return np.load(s)
        except ValueError:
            return np.loadtxt(s)
    else:
        return np.array([[float(i) for i in j.split(',')] for j in s.split()])


@click.group(chain=True, invoke_without_command=True)
@click.argument('out')
@click.option('--config', '-c', type=click.Path(exists=True),
              help="path of config file to load")
@click.option('--template/--no-template', is_eager=True,
              callback=clk.printconfigs,
              help="write default options to std out as config")
@click.option('--opts', '-opts', is_flag=True,
              help="check parsed options")
@click.option('--debug', is_flag=True,
              help="show traceback on exceptions")
@click.version_option(version=raytraverse.__version__)
@click.pass_context
def main(ctx, out, config, outconfig=None, **kwargs):
    """the raytraverse executable is a command line interface to the raytraverse
    python package for running and evaluating climate based daylight models.
    sub commands of raytraverse can be chained but should be invoked in the
    order given.

    the easiest way to manage options and sure that Scene and SunSetter classes
    are properly reloaded is to use a configuration file, to make a template::

        raytraverse --template > run.cfg

    after adjusting the settings, than each command can be invoked in turn and
    any dependencies will be loaded with the correct options, a complete run
    and evaluation can then be called by::

        raytraverse -c run.cfg OUT sky sunrun integrate

    as both scene and sun will be invoked automatically as needed.

    Arguments:
        * out: path to new or existing directory for raytraverse run
    """
    ctx.info_name = 'raytraverse'
    clk.get_config_chained(ctx, config, outconfig, None, None)
    ctx.obj = dict(out=out)


@main.command()
@click.option('-scene', help='space separated list of radiance scene files '
              '(no sky) or precompiled octree')
@click.option('-area', help='radiance scene file containing planar geometry of '
              'analysis area')
@click.option('-skyres', default=10.0,
              help='sky is subdivided accoring to a shirley-chiu disk to square'
                   ' mapping, total number of sky bins will equal skyres^2.'
                   ' solid angle of each patch will be 2*pi/(skyres^2)')
@click.option('-ptres', default=2.0,
              help='resolution of point subdivision on analysis plane. units'
                   ' match radiance scene file')
@click.option('-maxspec', default=0.3,
              help='an important parameter for guiding reflected sun rays.'
                   ' contribution values above this threshold are assumed to be'
                   ' direct view rays. If possible, (1) this value should be'
                   ' less than the tvis of the darkest glass in the scene, and'
                   ' (2) greater than the highest expected contribution from a'
                   ' specular reflection or scattering interaction. If it is'
                   ' not possible to meet both conditions, then ensure that'
                   ' condition (2) is met and consider using a substantially'
                   ' higher skyres to avoid massive over sampling of direct'
                   ' view rays')
@click.option('--reload/--no-reload', default=True,
              help='if a scene already exists at OUT reload it, note that if'
                   'this is False and overwrite is False, the program will'
                   'abort')
@click.option('--overwrite/--no-overwrite', default=False,
              help='Warning! if set to True and reload is False all files in'
                   'OUT will be deleted')
@click.option('--info/--no-info', default=False,
              help='print info on scene to stderr')
@click.option('--points/--no-points', default=False,
              help='print point locations to stdout')
@clk.shared_decs(clk.command_decs(raytraverse.__version__, wrap=True))
def scene(ctx, **kwargs):
    """The scene commands creates a Scene object which holds geometric
    information about the model including object geometry (and defined
    materials), the analysis plane and the desired resolutions for sky and
    analysis plane subdivision"""
    s = Scene(ctx.obj['out'], **kwargs)
    ctx.obj['scene'] = s
    if kwargs['points']:
        for pt in s.pts():
            print('{}\t{}\t{}'.format(*pt))
    if kwargs['info']:
        print(f'\nScene {s.outdir}:', file=sys.stderr)
        print('='*60 + '\n', file=sys.stderr)
        print('Scene Geometry:', file=sys.stderr)
        os.system(f'getinfo {s.scene}')
        print('Analysis Area:', file=sys.stderr)
        print('-'*60, file=sys.stderr)
        print(f'extents:\n{s.area.bbox}', file=sys.stderr)
        print(f'number of points: {s.area.npts}', file=sys.stderr)
        print(f'sky sampling resolution: {s.skyres}', file=sys.stderr)


run_opts = [
 click.option('-accuracy', default=1.0,
              help='a generic accuracy parameter that sets the threshold'
                   ' variance to sample. A value of 1 will have a sample count'
                   ' at the final sampling level equal to the number of'
                   ' directions with a contribution variance greater than .25'),
 click.option('-idres', default=4,
              help='the initial directional sampling resolution. each side'
                   ' of the sampling square (representing a hemisphere) will'
                   ' be subdivided 2^idres, yielding 2^(2*idres) samples and'
                   ' a resolution of 2^(2*idres)/(2pi) samples/steradian. this'
                   ' value should be smaller than 1/2 the size of the smallest'
                   ' view to an aperture that should be captured with 100%'
                   ' certainty'),
 click.option('--plotp/--no-plotp', default=False,
              help='for diagnostics only, plots the pdf at each level for'
                   ' point[0,0] in an interactive display (note that program'
                   ' will hang until the user closes the plot window at each'
                   ' level)'),
 click.option('--plotdview/--no-plotdview', default=False,
              help='plot a direct view of the sky field (as a .hdr file),'
                   ' this is equivalent to integrating with a value of 1 for'
                   ' all sky patches with no interpolation, plots pixels of'
                   ' actualsample vectors in red'),
 click.option('--run/--no-run', default=True,
              help='if True calls sampler.run()'),
 click.option('--rmraw/--no-rmraw', default=True,
              help='if True removes output of sampler.run(), after SCBinField'
                   ' is constructed. Note that SCBinField cannot be rebuilt'
                   ' once raw files are removed')
    ]


@main.command()
@click.option('-fdres', default=9,
              help='the final directional sampling resolution, yielding a'
                   ' grid of potential samples at 2^fdres x 2^fdres per'
                   ' hemisphere')
@click.option('-rcopts',
              default='-ab 7 -ad 60000 -as 30000 -lw 1e-7 -st 0 -ss 16',
              help='rtrace options to pass to the rcontrib call'
                   ' see the man pages for rtrace, rcontrib, and '
                   ' rcontrib -defaults for more information')
@clk.shared_decs(run_opts)
@clk.shared_decs(clk.command_decs(raytraverse.__version__, wrap=True))
def sky(ctx, plotdview=False, run=True, rmraw=True, executable='rcontrib',
        **kwargs):
    """the sky command intitializes and runs a sky sampler and then readies
    the results for integration by building a SCBinField. sky should be invoked
    before calling suns, as the sky contributions are used to select the
    necessary sun positions to run"""
    if 'scene' not in ctx.obj:
        clk.invoke_dependency(ctx, scene)
    sampler = SCBinSampler(ctx.obj['scene'], **kwargs)
    if run:
        sampler.run()
    sk = SCBinField(ctx.obj['scene'], rebuild=run, rmraw=rmraw)
    if plotdview:
        sk.direct_view()


@main.command()
@click.option('-srct', default=.01,
              help='if the contribution of a sky patch (for any view ray) is'
                   ' above this threshold, a sun will be created in this patch')
@click.option('-skyro', default=0.0,
              help='counter clockwise rotation (in degrees) of the sky to'
                   ' rotate true North to project North, so if project North'
                   ' is 10 degrees East of North, skyro=10')
@click.option('-sunres', default=10.0,
              help='resolution in degrees of the sky patch grid in which to'
                   ' stratify sun samples. Suns are randomly located within'
                   ' the grid, so this corresponds to the average distance'
                   ' between sources. The average error to a randomly selected'
                   ' sun position will be on average ~0.4 times this value')
@click.option('-loc', callback=clk.split_float,
              help='specify the scene location (if not specified in -wea or to'
                   ' override. give as "lat lon mer" where lat is + North, lon'
                   ' is + West and mer is the timezone meridian (full hours are'
                   ' 15 degree increments)')
@click.option('-wea',
              help="path to weather/sun position file. possible formats are:\n"
                   "\n"
                   "1. .wea file\n"
                   "#. .wea file without header (require -loc and "
                   "--no-usepositions)\n"
                   "#. .epw file\n"
                   "#. .epw file without header (require -loc and "
                   "--no-usepositions)\n"
                   "#. 3 column tsv file, each row is dx, dy, dz of candidate"
                   " sun position (requires --usepositions)\n"
                   "#. 4 column tsv file, each row is altitude, azimuth, direct"
                   " normal, diff. horizontal of canditate suns (requires "
                   "--usepositions)\n"
                   "#. 5 column tsv file, each row is dx, dy, dz, direct "
                   "normal, diff. horizontal of canditate suns (requires "
                   "--usepositions)\n\n"
                   "tsv files are loaded with loadtxt")
@click.option('--plotdview/--no-plotdview', default=False,
              help="creates a png showing sun positions on an angular fisheye"
                   " projection of the sky. sky patches are colored by the"
                   " maximum contributing ray to the scene")
@click.option('--reload/--no-reload', default=True,
              help="if False, regenerates sun positions, because positions may"
                   " be randomly selected this will make any sunrun results"
                   " obsolete")
@click.option('--usepositions/--no-usepositions', default=False,
              help='if True, sun positions will be chosen from the positions'
                   ' listed in wea. if more than one position is a candidate'
                   ' for that particular sky patch (as determined by sunres)'
                   ' than a random choice will be made. by using one of the'
                   ' tsv format options for wea, and preselecting sun positions'
                   ' such that there is 1 per patch a deterministic result can'
                   'be achieved.')
@clk.shared_decs(clk.command_decs(raytraverse.__version__, wrap=True))
def suns(ctx, loc=None, wea=None, usepositions=False, plotdview=False,
         **kwargs):
    """the suns command provides a number of options for creating sun positions
    used by sunrun see wea and usepositions options for details

    Note:

    the wea and skyro parameters are used to reduce the number of suns in cases
    where a specific site is known. Only suns within the solar transit (or
    positions if usepositions is True will be selected. It is important to note
    that when integrating, if a sun position outside this range is queried than
    results will not include the more detailed simulations involved in sunrun
    and will instead place the suns energy within the nearest sky patch. if
    skyres is small and or the patch is directly visible this will introduce
    significant bias in most metrics.
    """
    if 'scene' not in ctx.obj:
        clk.invoke_dependency(ctx, scene)
    if usepositions:
        if wea is None:
            raise ValueError('option -wea is required when use positions is '
                             'True')
        s = SunSetterPositions(ctx.obj['scene'], wea, **kwargs)
    elif loc is not None:
        s = SunSetterLoc(ctx.obj['scene'], loc, **kwargs)
    elif wea is not None:
        loc = raytraverse.skycalc.get_loc_epw(wea)
        s = SunSetterLoc(ctx.obj['scene'], loc, **kwargs)
    else:
        s = SunSetter(ctx.obj['scene'], **kwargs)
    if plotdview:
        s.direct_view()
    ctx.obj['suns'] = s


@main.command()
@click.option('-fdres', default=10,
              help='the final directional sampling resolution, yielding a'
                   ' grid of potential samples at 2^fdres x 2^fdres per'
                   ' hemisphere')
@click.option('-speclevel', default=9,
              help='at this sampling level, pdf is made from brightness of sky '
                   'sampling rather than progressive variance to look for fine '
                   'scale specular highlights, this should be atleast 1 level '
                   'from the end and the resolution of this level should be '
                   'smaller than the size of the source')
@click.option('-rcopts',
              default='-ab 6 -ad 3000 -as 1500 -st 0 -ss 16 -aa .1',
              help='rtrace options for sun reflection runs'
                   ' see the man pages for rtrace, and '
                   ' rtrace -defaults for more information')
@click.option('--view/--no-view', default=True,
              help="run/build/plot direct sun views")
@click.option('--ambcache/--no-ambcache', default=True,
              help='whether the rcopts indicate that the calculation will use '
                   'ambient caching (and thus should write an -af file argument'
                   ' to the engine)')
@click.option('--keepamb/--no-keepamb', default=False,
              help='whether to keep ambient files after run, if kept, a '
                   'successive call will load these ambient files, so care '
                   'must be taken to not change any parameters')
@click.option('--reflection/--no-reflection', default=True,
              help="run/build/plot reflected sun components")
@clk.shared_decs(run_opts)
@clk.shared_decs(clk.command_decs(raytraverse.__version__, wrap=True))
def sunrun(ctx, plotdview=False, run=True, rmraw=False, **kwargs):
    """the sunrun command intitializes and runs a sun sampler and then readies
    the results for integration by building a SunField."""
    if 'suns' not in ctx.obj:
        clk.invoke_dependency(ctx, suns)
    scn = ctx.obj['scene']
    sns = ctx.obj['suns']
    sampler = SunSampler(scn, sns, **kwargs)
    if run:
        sampler.run(kwargs['view'], kwargs['reflection'])
    if kwargs['view']:
        try:
            sv = SunViewField(scn, sns, rebuild=run, rmraw=rmraw)
        except FileNotFoundError as ex:
            print(f'Warning: {ex}', file=sys.stderr)
        else:
            if plotdview:
                sv.direct_view()
    if kwargs['reflection']:
        try:
            su = SunField(scn, sns, rebuild=run, rmraw=rmraw)
        except FileNotFoundError as ex:
            print(f'Warning: {ex}', file=sys.stderr)
        else:
            if plotdview:
                su.direct_view(res=200)


@main.command()
@click.option('-loc', callback=clk.split_float,
              help='specify the scene location (if not specified in -wea or to'
                   ' override. give as "lat lon mer" where lat is + North, lon'
                   ' is + West and mer is the timezone meridian (full hours are'
                   ' 15 degree increments)')
@click.option('-wea',
              help="path to weather/sun position file. possible formats are:\n"
                   "\n"
                   "1. .wea file\n"
                   "#. .wea file without header (requires -loc)\n"
                   "#. .epw file\n"
                   "#. .epw file without header (requires -loc)\n"
                   "#. 4 column tsv file, each row is altitude, azimuth, direct"
                   " normal, diff. horizontal of canditate suns (requires "
                   "--usepositions)\n"
                   "#. 5 column tsv file, each row is dx, dy, dz, direct "
                   "normal, diff. horizontal of canditate suns (requires "
                   "--usepositions)\n\n"
                   "tsv files are loaded with loadtxt"
              )
@click.option('-pts', default='0,0,0,0,-1,0', callback=np_load,
              help="points to evaluate, this can be a .npy file, a whitespace "
                   "seperated text file or entered as a string with commas "
                   "between components of a point and spaces between points. "
                   "in all cases each point requires 6 numbers x,y,z,dx,dy,dz "
                   "so the shape of the array will be (N, 6)")
@click.option('-res', default=800,
              help="the resolution of hdr output in pixels")
@click.option('-interp', default=12,
              help='number of nearby rays to use for interpolation of hdr'
                   'output (weighted by a gaussian filter). this does'
                   'not apply to metric calculations')
@click.option('-vname', default='view',
              help='name to include with hdr outputs')
@click.option('-skyro', default=0.0,
              help='counter clockwise rotation (in degrees) of the sky to'
                   ' rotate true North to project North, so if project North'
                   ' is 10 degrees East of North, skyro=10')
@click.option('--skyonly/--no-skyonly', default=False,
              help="if True, only integrate on Sky Field, useful for "
                   "diagnostics")
@click.option('--hdr/--no-hdr', default=True,
              help="produce an hdr output for each point and line in wea")
@click.option('--metric/--no-metric', default=True,
              help="calculate metrics for each point and wea file output"
                   " is ordered by point than sky")
@click.option('--header/--no-header', default=False,
              help='print column headings on metric output')
@clk.shared_decs(clk.command_decs(raytraverse.__version__, wrap=True))
def integrate(ctx, pts=None, skyonly=False, hdr=True,
              metric=True, res=800, interp=12, vname='view',
              header=False, **kwargs):
    """the integrate command combines sky and sun results and evaluates the
    given set of positions and sky conditions"""
    if 'scene' not in ctx.obj:
        clk.invoke_dependency(ctx, scene)
    scn = ctx.obj['scene']
    if not skyonly:
        if 'suns' not in ctx.obj:
            clk.invoke_dependency(ctx, suns)
        sns = ctx.obj['suns']
        su = SunField(scn, sns)
    else:
        su = None
    sk = SCBinField(scn)
    itg = Integrator(sk, su, **kwargs)
    skymtx = itg.get_sky_mtx()
    if hdr:
        itg.hdr(pts, *skymtx, interp=interp, res=res, vname=vname)
    if metric:
        mf = (raytraverse.metric.illum, raytraverse.metric.sqlum)
        metrics = itg.metric(pts, *skymtx, scale=179, metricfuncs=mf)
        if header:
            print("view\tpoint\tsky\t" + "\t".join([f.__name__ for f in mf]))
        for p, pts in enumerate(metrics):
            for s, skies in enumerate(pts):
                print(f"{p}\t{s}\t" + "\t".join([f"{i}" for i in skies]))


@main.resultcallback()
@click.pass_context
def printconfig(ctx, returnvalue, **kwargs):
    """callback to print additional info and cleanup any temp files"""
    rv = dict([i[0:2] for i in returnvalue])
    try:
        if 'scene' in rv:
            info = rv['scene']['info']
        else:
            info = (ctx.command.commands['scene'].
                    context_settings['default_map']['info'])
        if str(info).lower() in ('1', 'yes', 'y', 't', 'true'):
            s = ctx.obj['scene']
            print('\nCallback Scene Info:', file=sys.stderr)
            print('='*60 + '\n', file=sys.stderr)
            try:
                suncount = ctx.obj['suns'].suns.shape[0]
                print(f'scene has {suncount} suns', file=sys.stderr)
            except KeyError:
                print('sun setter not initialized', file=sys.stderr)
            print('\n Lightfield Data:', file=sys.stderr)
            print('-'*60, file=sys.stderr)
            print('Has sky lightfield data:',
                  os.path.isfile(f'{s.outdir}/sky_kd_data.pickle'),
                  file=sys.stderr)
            print('Has sunview lightfield data:',
                  os.path.isfile(f'{s.outdir}/sunview_kd_data.pickle'),
                  file=sys.stderr)
    except KeyError:
        pass
    try:
        clk.tmp_clean(ctx)
    except Exception:
        pass


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for raytraverse.scene"""
import os
import shutil

import pytest
from raytraverse.mapper import SpaceMapper
import numpy as np

# from clipt import mplt


@pytest.fixture(scope="module")
def tmpdir(tmp_path_factory):
    data = str(tmp_path_factory.mktemp("data"))
    shutil.copytree('tests/test/', data + '/test')
    cpath = os.getcwd()
    os.chdir(data + '/test')
    yield data + '/test'
    os.chdir(cpath)


def test_spacemapper(tmpdir):
    sm = SpaceMapper('plane.rad')
    assert np.all(np.equal(sm.pt2uv(sm.bbox), [[0, 0], [1, 1]]))
    sm2 = SpaceMapper('plane.rad', rotation=54)
    assert np.allclose(sm2.uv2pt(sm2._path[0].vertices), sm.uv2pt(sm._path[0].vertices))





# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import os

import numpy as np

from raytraverse import translate, quickplot, draw, renderer
from raytraverse.lightfield import SCBinField
from raytraverse.sampler.sampler import Sampler


class SingleSunSampler(Sampler):
    """sample contributions from direct suns.

    Parameters
    ----------
    scene: raytraverse.scene.Scene
        scene class containing geometry, location and analysis plane
    suns: raytraverse.sunsetter.SunSetter
        sun class containing sun locations.
    sidx: int
        sun index to sample
    speclevel: int, optional
        at this sampling level, pdf is made from brightness of sky sampling
        rather than progressive variance to look for fine scale specular
        highlights, this should be atleast 1 level from the end and the
        resolution of this level should be smaller than the size of the source
    keepamb: bool, optional
        whether to keep ambient files after run, if kept, a successive call
        will load these ambient files, so care must be taken to not change
        any parameters
    ambcache: bool, optional
        whether the rcopts indicate that the calculation will use ambient
        caching (and thus should write an -af file argument to the engine)
    """
    EngineClass = renderer.Rtrace

    def __init__(self, scene, suns, sidx, speclevel=9,
                 fdres=10, accuracy=1,
                 rcopts='-ab 6 -ad 3000 -as 1500 -st 0 -ss 16 -aa .1',
                 keepamb=False, ambcache=True, **kwargs):
        #: float: controls sampling limit in case of limited contribution
        self.slimit = suns.srct * .1
        self.srct = suns.srct
        anorm = accuracy * scene.skyres * (1 - np.cos(.533*np.pi/360))
        self.engine = self.EngineClass()
        engine_args = f"{rcopts} -oZ"

        # update ambient file and args before init
        self._keepamb = keepamb and ambcache
        if ambcache:
            engine_args += f" -af {scene.outdir}/sun_{sidx:04d}.amb"
        super().__init__(scene, stype=f"sun_{sidx:04d}", fdres=fdres,
                         accuracy=anorm, engine_args=engine_args,
                         srcdef=None, **kwargs)

        # update parameters post init
        #: int: index of level at which brightness sampling occurs
        self.specidx = speclevel - self.idres
        self.sidx = sidx
        #: np.array: sun position x,y,z
        self.sunpos = suns.suns[sidx]
        self.sbin = translate.uv2bin(translate.xyz2uv(self.sunpos[None, :],
                                                      flipu=False),
                                     self.scene.skyres).astype(int)[0]
        shape = np.concatenate((self.scene.area.ptshape, self.levels[0]))
        weights = self.pdf_from_sky(SCBinField(self.scene))
        self.weights = translate.resample(weights, shape)

        # load new source
        srcdef = f'{self.scene.outdir}/tmp_srcdef.rad'
        f = open(srcdef, 'w')
        f.write(suns.write_sun(sidx))
        f.close()
        self.engine.load_source(srcdef)
        os.remove(srcdef)

    def __del__(self):
        if not self._keepamb:
            try:
                os.remove(f'{self.scene.outdir}/sun_{self.sidx:04d}.amb')
            except (IOError, TypeError):
                pass
        super().__del__()

    def pdf_from_sky(self, skyfield, interp=12, rebuild=False, zero=True,
                     filterpts=True):
        ishape = np.concatenate((self.scene.area.ptshape,
                                 self.levels[self.specidx-2]))
        fi = f"{self.scene.outdir}/sunpdfidxs.npz"
        if os.path.isfile(fi) and not rebuild:
            f = np.load(fi)
            idxs = f['arr_0']
            errs = f['arr_1']
        else:
            shp = self.levels[self.specidx-2]
            si = np.stack(np.unravel_index(np.arange(np.product(shp)), shp))
            uv = (si.T + .5)/shp[1]
            grid = skyfield.scene.view.uv2xyz(uv)
            idxs, errs = skyfield.query_all_pts(grid, interp)
            strides = np.array(skyfield.lum.index_strides()[:-1])[:, None, None]
            idxs = np.reshape(idxs + strides, (-1, interp))
            errs = errs.reshape(-1, interp)
            np.savez(fi, idxs, errs)
        column = skyfield.lum.full_array()[:, self.sbin]
        if interp > 1:
            lum = np.average(column[idxs], -1, weights=1/errs).reshape(ishape)
        else:
            lum = column[idxs].reshape(ishape)
        if filterpts:
            haspeak = np.max(lum, (2, 3)) > self.srct
            lum = lum * haspeak[..., None, None]
        if zero:
            lum = np.where(lum > self.scene.maxspec, 0, lum)
        return lum

    def sample(self, vecf):
        """call rendering engine to sample sky contribution

        Parameters
        ----------
        vecf: str
            path of file name with sample vectors
            shape (N, 6) vectors in binary float format

        Returns
        -------
        lum: np.array
            array of shape (N,) to update weights
        """
        return super().sample(vecf).ravel()

    # @profile
    def draw(self):
        """draw samples based on detail calculated from weights
        detail is calculated across direction only as it is the most precise
        dimension

        Returns
        -------
        pdraws: np.array
            index array of flattened samples chosen to sample at next level
        """
        if self.idx == self.specidx:
            skyfield = SCBinField(self.scene)
            shape = np.concatenate((self.scene.area.ptshape,
                                    self.levels[self.idx]))
            weights = self.pdf_from_sky(skyfield)
            p = translate.resample(weights, shape)
            if self.plotp:
                quickplot.imshow(p[0, 0], [20, 10])
            pdraws = draw.from_pdf(p.ravel(), self.slimit)
        else:
            pdraws = super().draw()
        return pdraws



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import os
import shutil

from raytraverse.renderer.radiancerenderer import RadianceRenderer
from raytraverse.crenderer import cRcontrib
import raytraverse

path = os.path.dirname(raytraverse.__file__)
rcontrib_capture_file = f'{path}/tmp_rcontrib_capture'


class Rcontrib(RadianceRenderer):
    """singleton wrapper for c++ crenderer.cRcontrib singleton class"""
    Engine = cRcontrib
    name = 'rcontrib'
    arg_prefix = "-o !cat"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for raytraverse.craytraverse"""
import importlib
import sys

from raytraverse import craytraverse, io, renderer, draw
import numpy as np
from scipy import stats
import pytest

import clasp.script_tools as cst


def test_rtrace_call(capfd):
    args = "-ab 1 -ar 600 -ad 2000 -aa .2 -as 1500 -I"
    cargs = f"rtrace -h {args} -n 4 tests/test/test_run/sky.oct"
    check = cst.pipeline([cargs], inp='tests/test/rays.txt',
                         forceinpfile=True)
    check = np.fromstring(check, sep=' ').reshape(-1, 3)
    check2 = np.einsum('ij,j->i', check, [47.435/179, 119.93/179, 11.635/179])
    r = renderer.SPRtrace(args, "tests/test/test_run/sky.oct", iot='af')
    try:
        with capfd.disabled():
            ans = r.call('tests/test/rays.txt')
    except AttributeError:
        ans = r.call('tests/test/rays.txt')
    test = np.frombuffer(ans, '<f')
    r.initialize(args, "tests/test/test_run/sky.oct", iot='af')
    try:
        with capfd.disabled():
            ans = r.call('tests/test/rays.txt')
    except AttributeError:
        ans = r.call('tests/test/rays.txt')
    # test3 = np.fromstring(ans, sep=' ').reshape(-1, 3)
    test3 = np.frombuffer(ans, '<f')
    assert np.allclose(check2, test, atol=.03)
    assert np.allclose(check2, test3, atol=.03)
    # print(r.header)


def test_rcontrib_call(capfd):
    args = ('-V+ -I+ -ab 2 -ad 60000 -as 30000 -lw 1e-7 -e side:6'
            ' -f tests/test/scbins.cal -b bin -bn 36 -m skyglow ')
    cargs = f"rcontrib -n 5 -h- {args}  tests/test/test_run/sky.oct"
    check = cst.pipeline([cargs], inp='tests/test/rays2.txt',
                         forceinpfile=True)
    check = np.fromstring(check, sep=' ').reshape(-1, 36, 3)
    check = np.einsum('ikj,j->ik', check, [47.435/179, 119.93/179, 11.635/179])
    r = renderer.SPRcontrib(args, 'tests/test/test_run/sky.oct', iot='af')
    try:
        with capfd.disabled():
            ans = r.call('tests/test/rays.txt')
    except AttributeError:
        ans = r.call('tests/test/rays.txt')
    test = np.frombuffer(ans, '<f').reshape(-1, 36)
    assert np.allclose(check, test, atol=.03)




# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import shlex


class Renderer(object):
    """virtual renderer class"""

    initialized = False
    instance = None
    _pyinstance = None
    Engine = None
    name = None
    header = ""
    arg_prefix = ''

    def __new__(cls, rayargs=None, scene=None, nproc=None, **kwargs):
        if cls._pyinstance is None:
            cls._pyinstance = object.__new__(cls)
        return cls._pyinstance

    def __init__(self, rayargs=None, scene=None, nproc=None, **kwargs):
        self.initialize(rayargs, scene, nproc, **kwargs)

    @classmethod
    def _set_args(cls, args, iot, nproc):
        return shlex.split(f"{cls.name} -f{iot} -n {nproc} {cls.arg_prefix}"
                           f" {args}")

    @classmethod
    def initialize(cls, args, scene, nproc=None, **kwargs):
        pass

    @classmethod
    def call(cls, rayfile, store=True, outf=None):
        return None

    @classmethod
    def reset(cls):
        cls.initialized = False
        cls.header = ""

    @classmethod
    def reset_instance(cls):
        cls.instance = None
        cls.initialized = False
        cls.header = ""




# -*- coding: utf-8 -*-

# Copyright (c) 2019 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
from raytraverse.mapper.viewmapper import ViewMapper


class SunMapper(ViewMapper):
    """translate between view and normalized UV space

    Parameters
    ----------
    suns: np.array
        dx,dy,dz sun positions
    """

    def __init__(self, suns):
        super(SunMapper, self).__init__(suns, .533, name='sunmap')



# -*- coding: utf-8 -*-

# Copyright (c) 2019 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import numpy as np

from raytraverse.mapper.spacemapper import SpaceMapper


class SpaceMapperPt(SpaceMapper):
    """translate between world coordinates and normalized UV space"""

    @property
    def bbox(self):
        """np.array of shape (3,2): bounding box"""
        return self._bbox

    @property
    def sf(self):
        """bbox scale factor"""
        return self._sf

    @property
    def ptshape(self):
        """shape of point grid"""
        return self._ptshape

    @bbox.setter
    def bbox(self, plane):
        """read radiance geometry file as boundary path"""
        self._pts = np.loadtxt(plane)[:, 0:3]
        self._bbox = np.stack((np.min(self._pts, 0) - self.tolerance,
                               np.max(self._pts, 0) - self.tolerance))
        self._sf = len(self._pts)
        self._ptshape = np.array([self._sf, 1])

    def uv2pt(self, uv):
        """convert UV --> world

        Parameters
        ----------
        uv: np.array
            normalized UV coordinates of shape (N, 2)

        Returns
        -------
        pt: np.array
            world xyz coordinates of shape (N, 3)
        """

        idx = (uv[:, 0] * self.sf).astype(int)
        return self._pts[idx]

    def pt2uv(self, xyz):
        """convert world --> UV

        Parameters
        ----------
        xyz: np.array
            world xyz coordinates, shape (N, 3)

        Returns
        -------
        uv: np.array
            normalized UV coordinates of shape (N, 2)
        """
        perrs, pis = self.pt_kd.query(xyz)
        uv = np.full((xyz.shape[0], 2), .5)
        uv[:, 0] = (pis + .5) / self.sf
        return uv

    def idx2pt(self, idx):
        return self._pts[idx]

    def pts(self):
        return self._pts

    def in_area(self, xyz):
        """check if point is in boundary path

        Parameters
        ----------
        xyz: np.array
            uv coordinates, shape (N, 3)

        Returns
        -------
        mask: np.array
            boolean array, shape (N,)
        """
        perrs, pis = self.pt_kd.query(xyz)
        return perrs <= self.tolerance



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for raytraverse.craytraverse"""
import importlib
import sys

from raytraverse import craytraverse, io, renderer, draw
import numpy as np
from scipy import stats
import pytest

import clasp.script_tools as cst


def test_from_pdf():
    rv = stats.norm()
    nsamp = 500000
    t = .01
    x = np.linspace(0, 5, nsamp)
    pdf = rv.pdf(x)
    exp = np.sum(pdf > t)
    # c2 = draw.from_pdf2(pdf, t)
    c = draw.from_pdf(pdf, t)
    # print(np.sum(pdf > t), c.size, c2.size)
    # hist = np.histogram(x[c], 50)
    # mplt.quick_scatter([hist[1][1:], [0, 0]], [hist[0], [0, 7000]])
    # print(exp, c.size, c2.size)
    assert np.isclose(c.size, exp)


def test_empty_reset():
    rt = renderer.Rtrace()
    rt.reset()
    rt.reset()
    rt.reset_instance()
    rt = renderer.Rcontrib()
    rt.reset()
    rt.reset()
    rt.reset_instance()
    assert True


def test_rtrace_call(capfd):
    args = "-ab 1 -ar 600 -ad 2000 -aa .2 -as 1500 -I"
    cargs = f"rtrace -h {args} -n 4 tests/test/test_run/sky.oct"
    check = cst.pipeline([cargs], inp='tests/test/rays.txt',
                         forceinpfile=True)
    check = np.fromstring(check, sep=' ').reshape(-1, 3)
    check2 = np.einsum('ij,j->i', check, [47.435/179, 119.93/179, 11.635/179])
    r = renderer.Rtrace(args, "tests/test/test_run/sky.oct", iot='aa')
    print('call')
    try:
        with capfd.disabled():
            ans = r.call('tests/test/rays.txt')
    except AttributeError:
        ans = r.call('tests/test/rays.txt')
    print('call done')
    test = np.fromstring(ans, sep=' ').reshape(-1, 3)
    r.update_ospec('ZL', 'a')
    try:
        with capfd.disabled():
            ans = r.call('tests/test/rays.txt')
    except AttributeError:
        ans = r.call('tests/test/rays.txt')
    test2 = np.fromstring(ans, sep=' ').reshape(-1, 2)
    r.reset()
    args2 = args + ' -oZ'
    r.initialize(args2, "tests/test/test_run/sky.oct", iot='af')
    try:
        with capfd.disabled():
            ans = r.call('tests/test/rays.txt')
    except AttributeError:
        ans = r.call('tests/test/rays.txt')
    # test3 = np.fromstring(ans, sep=' ').reshape(-1, 3)
    test3 = np.frombuffer(ans, '<f')
    assert np.allclose(check, test, atol=.03)
    assert np.allclose(check2, test2[:, 0], atol=.03)
    assert np.allclose(check2, test3, atol=.03)
    # print(r.header)


def test_rcontrib_call(capfd):
    args = ('-V+ -I+ -ab 2 -ad 60000 -as 30000 -lw 1e-7 -e side:6'
            ' -f tests/test/scbins.cal -b bin -bn 36 -m skyglow ')
    cargs = f"rcontrib -n 5 -h- {args}  tests/test/test_run/sky.oct"
    check = cst.pipeline([cargs], inp='tests/test/rays2.txt',
                         forceinpfile=True)
    check = np.fromstring(check, sep=' ').reshape(-1, 36, 3)
    check = np.einsum('ikj,j->ik', check, [47.435/179, 119.93/179, 11.635/179])
    r = renderer.Rcontrib('-Z+' + args, 'tests/test/test_run/sky.oct', iot='aa')
    try:
        with capfd.disabled():
            ans = r.call('tests/test/rays.txt')
    except AttributeError:
        ans = r.call('tests/test/rays.txt')
    test = np.fromstring(ans, sep=' ').reshape(-1, 36)
    assert np.allclose(check, test, atol=.03)


if __name__ == "__main__":
    # test_rcontrib_call()
    rc = renderer.Rcontrib()
    # rt = renderer.Rtrace()
    # print(rt.instance)
    test_rtrace_call(None)
    rt = renderer.Rtrace()
    print(rt.header)
    rt.reset()
    test_rtrace_call(None)
    rt.reset_instance()
    test_rtrace_call(None)
    rt = renderer.Rtrace()
    rt.reset_instance()
    # print(rt.instance)
    print('rt done')

    # print(rc.instance)
    print('rc 1')
    test_rcontrib_call(None)
    print(rc.header)
    rc.reset()
    # print(rc.instance)
    print('rc 2')
    test_rcontrib_call(None)
    # rc.reset_instance()
    print('rc 3')
    test_rcontrib_call(None)
    rc = renderer.Rcontrib()
    # print(rc.instance)
    print('rc 4')
    test_rcontrib_call(None)
    test_rtrace_call(None)




# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import os
import pickle
import sys
from concurrent.futures import ThreadPoolExecutor, as_completed
from concurrent.futures import ProcessPoolExecutor
from scipy.stats import norm
from scipy.spatial import cKDTree, SphericalVoronoi, _voronoi

import numpy as np
from raytraverse.lightfield.memarraydict import MemArrayDict

from raytraverse import io, translate
from raytraverse.lightfield.lightfield import LightField


class LightFieldKD(LightField):
    """light field with KDtree structures for spatial query"""

    @staticmethod
    def mk_vector_ball(v):

        class SVoronoi(SphericalVoronoi):
            """this is a temporary fix for an apperent bug in
            SphericalVoronoi"""
            def sort_vertices_of_regions(self):
                if self._dim != 3:
                    raise TypeError(
                        "Only supported for three-dimensional point sets")
                reg = [r for r in self.regions if len(r) > 0]
                _voronoi.sort_vertices_of_regions(self._simplices, reg)

        d_kd = cKDTree(v)
        omega = SVoronoi(v).calculate_areas()[:, None]
        return d_kd, omega

    @property
    def d_kd(self):
        """list of direction kdtrees

        :getter: Returns kd tree structure
        :type: list of scipy.spatial.cKDTree
        """
        return self._d_kd

    @property
    def scene(self):
        """scene information

        :getter: Returns this integrator's scene
        :setter: Set this integrator's scene
        :type: raytraverse.scene.Scene
        """
        return self._scene

    @scene.setter
    def scene(self, scene):
        """Set this field's scene and load samples"""
        self._scene = scene
        kdfile = f'{scene.outdir}/{self.prefix}_kd_data.pickle'
        lumfile = f'{scene.outdir}/{self.prefix}_kd_lum_map.pickle'
        lumdat = f'{self.scene.outdir}/{self.prefix}_kd_lum.dat'
        if (os.path.isfile(kdfile) and os.path.isfile(lumfile)
                and not self.rebuild):
            f = open(kdfile, 'rb')
            self._d_kd = pickle.load(f)
            self._vec = pickle.load(f)
            self._omega = pickle.load(f)
            f.close()
            f = open(lumfile, 'rb')
            self._lum = pickle.load(f)
            f.close()
        else:
            f = open(lumdat, 'wb')
            f.close()
            self._d_kd, self._vec, self._omega, self._lum = self._mk_tree()
            f = open(kdfile, 'wb')
            pickle.dump(self.d_kd, f, protocol=4)
            pickle.dump(self.vec, f, protocol=4)
            pickle.dump(self.omega, f, protocol=4)
            f.close()
            f = open(lumfile, 'wb')
            pickle.dump(self._lum, f, protocol=4)
            f.close()

    def raw_files(self):
        """get list of files used to build field"""
        dfile = f'{self.scene.outdir}/{self.prefix}_vals.out'
        vfile = f'{self.scene.outdir}/{self.prefix}_vecs.out'
        return [dfile, vfile]

    def _to_mem_map(self, ar, offset=0):
        outf = f'{self.scene.outdir}/{self.prefix}_kd_lum.dat'
        mar = np.memmap(outf, dtype='<f', mode='r+',
                        offset=offset, shape=ar.shape)
        mar[:] = ar[:]
        mar.flush()
        del mar
        return outf, '<f', 'r', offset, ar.shape

    def _get_vl(self, npts, pref='', ltype=MemArrayDict, os0=0, fvrays=False):
        dfile = f'{self.scene.outdir}/{self.prefix}{pref}_vals.out'
        vfile = f'{self.scene.outdir}/{self.prefix}{pref}_vecs.out'
        if not (os.path.isfile(dfile) and os.path.isfile(vfile)):
            raise FileNotFoundError("No results files found, have you run"
                                    f" a Sampler of type {self.prefix} for"
                                    f" scene {self.scene.outdir}?")
        fvecs = io.bytefile2np(open(vfile, 'rb'), (-1, 4))
        alums = io.bytefile2np(open(dfile, 'rb'), (fvecs.shape[0], self.srcn))
        if fvrays:
            blindsquirrel = (np.max(alums, 1) < self.scene.maxspec)
            fvecs = fvecs[blindsquirrel]
            alums = alums[blindsquirrel]
        sorting = fvecs[:, 0].argsort()
        fvecs = fvecs[sorting]
        pidx = fvecs[:, 0]
        pt_div = np.searchsorted(pidx, np.arange(npts), side='right')
        pt0 = 0
        vecs = {}
        lums = []
        lummap = []
        with ThreadPoolExecutor() as exc:
            pts = []
            for i, pt in enumerate(pt_div):
                if pt != pt0:
                    pts.append(i)
                    vecs[i] = fvecs[pt0:pt, 1:4]
                    slc = sorting[pt0:pt]
                    ar = alums[slc]
                    lums.append(exc.submit(self._to_mem_map, ar,
                                           offset=os0 + 4*pt0*self.srcn))
                    pt0 = pt
            for pt, lm in zip(pts, lums):
                lummap.append((pt, lm.result()))
        return vecs, ltype(lummap)

    def _mk_tree(self, pref='', ltype=MemArrayDict, os0=0):
        npts = self.scene.area.npts
        vs, lums = self._get_vl(npts, pref=pref, ltype=ltype, os0=os0)
        with ProcessPoolExecutor() as exc:
            d_kd, omega = zip(*exc.map(LightFieldKD.mk_vector_ball,
                                       vs.values()))
        return dict(zip(vs.keys(), d_kd)), vs, dict(zip(vs.keys(), omega)), lums

    def apply_coef(self, pi, coefs):
        c = np.asarray(coefs).reshape(-1, 1)
        return np.einsum('ij,kj->ik', c, self.lum[pi])

    def add_to_img(self, img, mask, pi, i, d, coefs=1, vm=None, radius=3):
        lum = self.apply_coef(pi, coefs)
        if len(i.shape) > 1:
            # gaussian reconstruction filter
            y = norm(scale=translate.theta2chord(radius*np.pi/180))
            w = np.broadcast_to(y.pdf(d), (lum.shape[0],) + d.shape)
            lum = np.average(lum[:, i], weights=w, axis=-1)
        else:
            lum = lum[:, i]
        img[mask] += np.squeeze(lum)

    def query_ray(self, pi, vecs, interp=1):
        d, i = self.d_kd[pi].query(vecs, k=interp)
        return i, d

    def query_all_pts(self, vecs, interp=1):
        futures = []
        idxs = []
        errs = []
        with ProcessPoolExecutor() as exc:
            for pt in self.items():
                futures.append(exc.submit(self.d_kd[pt].query, vecs, interp))
            for fu in futures:
                r = fu.result()
                idxs.append(r[1])
                errs.append(r[0])
        return np.stack(idxs), np.stack(errs)

    def query_ball(self, pi, vecs, viewangle=180):
        vs = translate.theta2chord(viewangle/360*np.pi)
        return self.d_kd[pi].query_ball_point(translate.norm(vecs), vs)

    def _dview(self, idx, pdirs, mask, res=800, showsample=True):
        img = np.zeros((res*self.scene.view.aspect, res))
        i, d = self.query_ray(idx, pdirs[mask])
        self.add_to_img(img, mask, idx, i, d)
        outf = f"{self.outfile(idx)}.hdr"
        if showsample:
            vm = self.scene.view
            img = np.repeat(img[None, ...], 3, 0)
            vi = self.query_ball(idx, vm.dxyz, vm.viewangle)
            v = self.vec[idx][vi[0]]
            reverse = vm.degrees(v) > 90
            pa = vm.ivm.ray2pixel(v[reverse], res)
            pa[:, 0] += res
            pb = vm.ray2pixel(v[np.logical_not(reverse)], res)
            xp = np.concatenate((pa[:, 0], pb[:, 0]))
            yp = np.concatenate((pa[:, 1], pb[:, 1]))
            img[1:, xp, yp] = 0
            img[0, xp, yp] = 1
            io.carray2hdr(img, outf)
        else:
            io.array2hdr(img, outf)
        return outf

    def direct_view(self, res=800, showsample=True, items=None):
        """create a summary image of lightfield for each vpt"""
        vm = self.scene.view
        pdirs = vm.pixelrays(res)
        if items is None:
            items = self.items()
        if vm.aspect == 2:
            mask = vm.in_view(np.concatenate((pdirs[0:res],
                                              -pdirs[res:]), 0))
        else:
            mask = vm.in_view(pdirs)
        fu = []
        with ThreadPoolExecutor() as exc:
            for idx in items:
                fu.append(exc.submit(self._dview, idx, pdirs, mask, res,
                                     showsample))
            for f in as_completed(fu):
                print(f.result(), file=sys.stderr)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for raytraverse.io"""
import os
import shutil
import sys

import pytest
from raytraverse import io
import numpy as np


@pytest.fixture(scope="module")
def tmpdir(tmp_path_factory):
    data = str(tmp_path_factory.mktemp("data"))
    cpath = os.getcwd()
    os.chdir(data)
    yield data
    os.chdir(cpath)


def test_array2img(tmpdir):
    b, a = np.mgrid[0:600, 0:400]
    ar = a*b
    io.array2hdr(ar, 'mgrid.hdr')
    io.array2hdr(a, 'mgrida.hdr')
    io.array2hdr(b, 'mgridb.hdr')
    a2 = io.hdr2array('mgrida.hdr')
    b2 = io.hdr2array('mgridb.hdr')
    ar2 = io.hdr2array('mgrid.hdr')
    assert np.allclose(a.T, a2, atol=.25, rtol=.03)
    assert np.allclose(b.T, b2, atol=.25, rtol=.03)
    assert np.allclose(ar2, ar2, atol=.25, rtol=.03)


def test_capture(tmpdir):
    inp = ("cnt 100 | rcalc -f rayinit.cal -f b2d.cal -e 'side=10;bin=$1"
           ";$1=$1;$2=U;$3=V;$4=Dx;$5=Dy;$6=Dz' | rcalc -f rayinit.cal -f "
           "d2b.cal -e 'side=10;Dx=$4;Dy=$5;Dz=$6;$1=bin'")
    f = open("test.txt", 'w')
    with io.CaptureStdOut(outf=f) as cap:
        pass
    with io.CaptureStdOut(outf=f) as cap:
        print(inp)
    f.close()
    f = open('test.txt').read()
    assert inp + "\n" == f
    assert cap.stdout == inp + "\n"
    inp = np.arange(10)
    f = open("test.txt", 'wb')
    with io.CaptureStdOut(b=True, outf=f) as cap:
        sys.stdout.buffer.write(io.np2bytes(inp))
    f.close()
    f = open('test.txt', 'rb').read()
    assert np.allclose(inp, io.bytes2np(f, (10,)))
    assert np.allclose(inp, io.bytes2np(cap.stdout, (10,)))
    with pytest.raises(AttributeError):
        with io.CaptureStdOut(b=True, outf="test.txt"):
            pass



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import os
import pickle

import numpy as np

from raytraverse import draw, renderer
from raytraverse.lightfield import SCBinField
from raytraverse.sampler import Sampler


class SunViewSampler(Sampler):
    """sample view rays to direct suns.

    here idres and fdres are sampled on a per sun basis for a view centered
    on each sun direction with a view angle of .533 degrees (hardcoded in
    sunmapper class).

    Parameters
    ----------
    scene: raytraverse.scene.Scene
        scene class containing geometry, location and analysis plane
    suns: raytraverse.sunsetter.SunSetter
        sun class containing sun locations.
    loadsrc: bool
        include suns.rad in base scene initialization. if False,
        self.engine.load_source must be invoked before call.
    """
    EngineClass = renderer.Rtrace

    def __init__(self, scene, suns, **kwargs):
        self.suns = suns
        self.engine = self.EngineClass()
        super().__init__(scene, stype='sunview', idres=4, fdres=6,
                         srcdef=None, engine_args='-oZ -ab 0', **kwargs)
        self.engine.load_source(f"{scene.outdir}/suns.rad")
        self.samplemap = self.suns.map

    @property
    def levels(self):
        """sampling scheme

        :getter: Returns the sampling scheme
        :setter: Set the sampling scheme from (ptres, fdres, skres)
        :type: np.array
        """
        return self._levels

    @levels.setter
    def levels(self, fdres):
        """calculate sampling scheme"""
        self._levels = np.array([(self.suns.suns.shape[0], 2**i, 2**i)
                                 for i in range(self.idres, fdres + 1, 1)])

    def check_viz(self):
        skyfield = SCBinField(self.scene)
        idx, errs = skyfield.query_all_pts(self.suns.suns, 4)
        isviz = np.array([np.max(skyfield.lum[i][j], (1, 2)) > self.suns.srct/2
                          for i, j in enumerate(idx)])
        isviz = isviz.reshape(*self.scene.area.ptshape,  -1)
        suns = np.broadcast_to(isviz[..., None, None].astype(int),
                               isviz.shape + (self.weights.shape[-2:]))
        return suns

    def sample(self, vecf):
        """call rendering engine to sample direct view rays

        Parameters
        ----------
        vecf: str
            path of file name with sample vectors
            shape (N, 6) vectors in binary float format

        Returns
        -------
        lum: np.array
            array of shape (N,) to update weights
        """
        return super().sample(vecf).ravel()

    def _uv2xyz(self, uv, si):
        return self.samplemap.uv2xyz(uv, si[2])

    def draw(self):
        """draw first level based on sky visibility"""
        if self.idx == 0:
            p = self.check_viz().ravel()
            pdraws = draw.from_pdf(p, .5)
        else:
            pdraws = super().draw()
        return pdraws

    def run_callback(self):
        """post sampling, right full resolution (including interpolated values)
         non zero rays to result file."""
        super().run_callback()
        shape = self.levels[self.idx, -2:]
        size = np.prod(shape)
        vals = self.weights.reshape(-1, self.weights.shape[2], size)
        si = np.stack(np.unravel_index(np.arange(size), shape)).T
        uv = ((si + .5)/shape)
        vecs = []
        lums = []
        for i in range(vals.shape[0]):
            ptv = []
            ptl = []
            for j in range(vals.shape[1]):
                valid = vals[i, j] > self.suns.srct
                cnt = np.sum(valid)
                if cnt > 0:
                    ptv.append(uv[valid])
                    ptl.append(vals[i, j][valid])
                else:
                    ptv.append(np.arange(0))
                    ptl.append(np.arange(0))
            lums.append(ptl)
            vecs.append(ptv)
        outf = f'{self.scene.outdir}/{self.stype}_result.pickle'
        f = open(outf, 'wb')
        pickle.dump(vecs, f, protocol=4)
        pickle.dump(lums, f, protocol=4)
        pickle.dump(shape, f, protocol=4)
        f.close()
        os.remove(f'{self.scene.outdir}/{self.stype}_vals.out')



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

"""sampler objects"""

__all__ = ['Sampler', 'SCBinSampler', 'SunViewSampler',
           'SunSampler', 'SingleSunSampler']

from raytraverse.sampler.sampler import Sampler
from raytraverse.sampler.scbinsampler import SCBinSampler
from raytraverse.sampler.singlesunsampler import SingleSunSampler
from raytraverse.sampler.sunviewsampler import SunViewSampler
from raytraverse.sampler.sunsampler import SunSampler




# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import itertools
import os
from concurrent.futures import ProcessPoolExecutor

import numpy as np
from scipy.stats import norm

from clasp.script_tools import pipeline, sglob
from raytraverse.lightfield.memarraydict import MemArrayDict
from raytraverse import translate
from raytraverse.lightfield.lightfieldkd import LightFieldKD
from raytraverse.lightfield.sunviewfield import SunViewField


class SunField(LightFieldKD):
    """container for sun view data

    Parameters
    ----------
    scene: raytraverse.scene.Scene
        scene class containing geometry, location and analysis plane
    suns: raytraverse.sunsetter.SunSetter
        sun class containing sun vectors and SunMapper (passed to SunViewField)
    rebuild: bool, optional
        build kd-tree even if one exists
    """

    def __init__(self, scene, suns, rebuild=False, rmraw=False):
        #: raytraverse.sunsetter.SunSetter
        self.suns = suns
        #: raytraverse.lightfield.SunViewField
        self.view = SunViewField(scene, suns, rebuild=False, rmraw=rmraw)
        super().__init__(scene, rebuild=rebuild, prefix='sun', rmraw=rmraw)

    def raw_files(self):
        """get list of files used to build field"""
        rf = []
        for i in range(self.suns.suns.shape[0]):
            rf.append(f'{self.scene.outdir}/{self.prefix}_{i:04d}_vals.out')
            rf.append(f'{self.scene.outdir}/{self.prefix}_{i:04d}_vecs.out')
        return rf + self.view.raw_files()

    def _mk_tree(self, pref='', ltype=list, os0=0):
        d_kds = {}
        vecs = {}
        omegas = {}
        lums = MemArrayDict({})
        offset = 0
        npts = self.scene.area.npts
        with ProcessPoolExecutor() as exc:
            futures = []
            for i in range(self.suns.suns.shape[0]):
                dfile = f'{self.scene.outdir}/{self.prefix}_{i:04d}_vals.out'
                if os.path.isfile(dfile):
                    vs, lum = self._get_vl(npts, pref=f'_{i:04d}', ltype=ltype,
                                           os0=offset, fvrays=True)
                    lasti = lum[-1][1]
                    offset = lasti[-2] + lasti[-1][0]*lasti[-1][1]*4
                    for j, lm in lum:
                        vecs[(j, i)] = vs[j]
                        lums[(j, i)] = lm
                        futures.append(((j, i),
                                       exc.submit(LightFieldKD.mk_vector_ball,
                                       vs[j])))
            for fu in futures:
                idx = fu[0]
                d_kds[idx], omegas[idx] = fu[1].result()
        return d_kds, vecs, omegas, lums

    def items(self):
        return self.d_kd.keys()

    def keymap(self):
        npts = self.scene.area.npts
        shape = (npts, self.suns.suns.shape[0] + 1)
        idxgrid = np.unravel_index(np.arange(shape[1] * npts), shape)
        full = np.core.records.fromarrays(idxgrid)
        items = np.core.records.fromarrays(list(zip(*self.items())))
        return np.isin(full, items).reshape(shape)

    def add_to_img(self, img, mask, pi, i, d, coefs=1, vm=None, radius=3):
        if vm is None:
            vm = self.scene.view
        lum = self.apply_coef(pi, coefs)
        if len(i.shape) > 1:
            y = norm(scale=translate.theta2chord(radius*np.pi/180))
            w = np.broadcast_to(y.pdf(d), (lum.shape[0],) + d.shape)
            lum = np.average(lum[:, i], weights=w, axis=-1)
        else:
            lum = lum[:, i]
        img[mask] += np.squeeze(lum)
        sun = np.concatenate((self.suns.suns[pi[1]], [coefs, ]))
        self.view.add_to_img(img, pi, sun, vm)

    def direct_view(self, res=200, showsample=False, items=None):
        """create a summary image of lightfield for each vpt"""
        super().direct_view(res=res, showsample=showsample, items=items)
        if items is not None:
            return None
        for i in super().items():
            flist = sglob(f"{self.scene.outdir}_{self.prefix}_{i:04d}_*.hdr")
            ssq = int(np.ceil(np.sqrt(len(flist))))
            files = ' '.join(flist)
            outf = f"{self.scene.outdir}_{self.prefix}_{i:04d}.hdr"
            pcompos = f'pcompos -a -{ssq} -s 5 -b 1 1 1 -la {files}'
            pipeline([pcompos], outf, close=True, writemode='wb')
            for fl in flist:
                os.remove(fl)



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import numpy as np

from raytraverse import translate
from raytraverse.scene.skyinfo import SkyInfo
from raytraverse.scene.sunsetter import SunSetter


class SunSetterLoc(SunSetter):
    """select suns to sample based on sky pdf and scene.

    Parameters
    ----------
    scene: raytraverse.scene.Scene
        scene class containing geometry, location and analysis plane
    loc: tuple
        lat, lon, tz (in degrees, west is positive
    srct: float, optional
        threshold of sky contribution for determining appropriate srcn
    skyro: float, optional
        sky rotation (in degrees, ccw)
    reload: bool
        if True reloads existing sun positions, else always generates new
    """

    def __init__(self, scene, loc, skyro=0.0, **kwargs):
        #: raytraverse.scene.SkyInfo
        self.sky = SkyInfo(loc, skyro)
        super().__init__(scene, skyro=skyro, **kwargs)

    def choose_suns(self):
        uvsize = self.sunres
        si = np.stack(np.unravel_index(np.arange(uvsize**2),
                                       (uvsize, uvsize)))
        skyb = self.load_sky_facs()
        uv = si.T/uvsize
        ib = self.sky.in_solarbounds(uv + .5/uvsize,
                                     size=1/uvsize)
        ib = (skyb*ib > self.srct)
        border = 2*uvsize/180
        j = np.random.default_rng().uniform(border, 1 - border, si.T[ib].shape)
        # j = .5
        uv = (si.T[ib] + j)/uvsize
        xyz = translate.uv2xyz(uv, xsign=1)
        return xyz



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for raytraverse.scene"""
import os
import shutil

import pytest
from click.testing import CliRunner
import numpy as np
import clasp.click_ext as clk

from raytraverse import cli
import hdrstats.cli as hdrcli
from hdrstats.hdrstats import corr_calc


@pytest.fixture()
def runner():
    return CliRunner(mix_stderr=False)


@pytest.fixture(scope="module")
def tmpdir(tmp_path_factory):
    data = str(tmp_path_factory.mktemp("data"))
    shutil.copytree('tests/workflow/', data + '/workflow')
    cpath = os.getcwd()
    os.chdir(data + '/workflow')
    yield data + '/workflow'
    os.chdir(cpath)


@pytest.mark.skip('slow')
def test_cli(tmpdir, capfd, runner):
    with pytest.raises(SystemExit) as exc_info:
        with capfd.disabled():
            cli.main.main(args=["-c", "run.cfg", "demo", "sky", "sunrun"])
    assert exc_info.value.args[0] == 0
    runner.invoke(cli.main, "-c run.cfg demo integrate --no-metric")
    hdr = runner.invoke(hdrcli.img_cr, "'demo_view*.hdr'")
    hdr = np.fromstring(hdr.output, sep=' ').reshape(-1, 5)[:, 1:3]
    pt = runner.invoke(cli.main, "-c run.cfg demo integrate --no-hdr")
    pt = np.fromstring(pt.output, sep=' ').reshape(-1, 5)[:, 2:4]
    corr = corr_calc(hdr[:, 0], pt[:, 0])
    assert corr[0] > .95



# -*- coding: utf-8 -*-

# Copyright (c) 2019 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

"""functions for reading and writing"""
import shlex
import os
import sys
import threading
from subprocess import Popen, PIPE
from io import StringIO, BytesIO

import numpy as np


class CaptureStdOut:
    """redirect output streams at system level (including c printf)

    Parameters
    ----------

    b: bool, optional
        read data as bytes
    store: bool, optional
        record stdout in a IOStream, value accesible through self.stdout
    outf: IOBase, optional
        if not None, must be writable, closed on exit

    Notes
    -----
    ::

        with CaptureStdOut() as capture:
            do stuff
        capout = capture.stdout

    when using with pytest include the -s flag or this class has no effect

    """

    def __init__(self, b=False, store=True, outf=None):
        # Create pipe and dup2() the write end of it on top of stdout,
        # saving a copy of the old stdout
        if outf is not None and not hasattr(outf, "write"):
            raise AttributeError('If outf is not None, it must have a write'
                                 ' attribute')
        self.fileno = sys.stdout.fileno()
        self.save = os.dup(self.fileno)
        self.pipe = os.pipe()
        os.dup2(self.pipe[1], self.fileno)
        os.close(self.pipe[1])
        self._stdout = None
        self._file = outf
        if b:
            self.threader = threading.Thread(target=self.drain_bytes)
            if store:
                self._stdout = BytesIO()
        else:
            self.threader = threading.Thread(target=self.drain_str)
            if store:
                self._stdout = StringIO()
        self.threader.start()

    @property
    def stdout(self):
        try:
            return self._stdout.getvalue()
        except AttributeError:
            return None

    def __enter__(self):
        return self

    def _write(self, data):
        try:
            self._stdout.write(data)
        except AttributeError:
            pass
        try:
            self._file.write(data)
        except AttributeError:
            pass

    def drain_bytes(self):
        """read stdout as bytes"""
        while True:
            data = os.read(self.pipe[0], 1024)
            if not data:
                break
            self._write(data)

    def drain_str(self):
        """read stdout as unicode"""
        while True:
            data = os.read(self.pipe[0], 1024).decode()
            if not data:
                break
            self._write(data)

    def __exit__(self, exc_type, exc_value, traceback):
        """restore stdout and join threads"""
        # Close the write end of the pipe to unblock the reader thread and
        # trigger it to exit
        os.close(self.fileno)
        self.threader.join()
        # Clean up the pipe and restore the original stdout
        os.close(self.pipe[0])
        os.dup2(self.save, self.fileno)
        os.close(self.save)


def call_sampler(outf, command, vecs, shape):
    """make subprocess call to sampler given as command, expects rgb value
    as return for each vec

    Parameters
    ----------
    outf: str
        path to write out to
    command: str
        command line with executable and options
    vecs: np.array
        vectors to pass as stdin to command
    shape: tuple
        shape of expected output

    Returns
    -------
    lums: np.array
        of length vectors.shape[0]

    """
    f = open(outf, 'a+b')
    lum_file_pos = f.tell()
    p = Popen(shlex.split(command), stdout=f, stdin=PIPE)
    p.communicate(np2bytes(vecs))
    lum = bytefile2rad(f, shape, subs='ijk,k->ij', offset=lum_file_pos)
    return lum


def bytefile2rad(f, shape, slc=..., subs='ijk,k->ij', offset=0):
    memarray = np.memmap(f, dtype='<f', mode='r', shape=shape, offset=offset)
    return np.einsum(subs, memarray[slc], [0.265, 0.670, 0.065])


def np2bytes(ar, dtype='<f'):
    """format ar as bytestring

    Parameters
    ----------
    ar: np.array
    dtype: str
        argument to pass to np.dtype()

    Returns
    -------
    bytes
    """
    dt = np.dtype(dtype)
    return ar.astype(dt).tobytes()


def bytes2np(buf, shape, dtype='<f'):
    """read ar from bytestring

    Parameters
    ----------
    buf: bytes, str
    shape: tuple
        array shape
    dtype: str
        argument to pass to np.dtype()

    Returns
    -------
    np.array
    """
    dt = np.dtype(dtype)
    return np.frombuffer(buf, dtype=dt).reshape(*shape)


def bytefile2np(f, shape, dtype='<f'):
    """read binary data from f

    Parameters
    ----------
    f: IOBase
        file object to read array from
    shape: tuple
        array shape
    dtype: str
        argument to pass to np.dtype()

    Returns
    -------
    ar.shape
        necessary for reconstruction
    """
    return bytes2np(f.read(), shape, dtype)


def array2hdr(ar, imgf, header=None):
    """write 2d np.array (x,y) to hdr image format

    Parameters
    ----------
    ar: np.array
    imgf: file path to right
    header: list of header lines to append to image header

    Returns
    -------

    """
    f = open(imgf, 'wb')
    pval = f'pvalue -r -b -h -H -df -o -y {ar.shape[1]} +x {ar.shape[0]}'
    if header is not None:
        hdr = "' '".join(header)
        getinfo = shlex.split(f"getinfo -a '{hdr}'")
        p = Popen(pval.split(), stdin=PIPE, stdout=PIPE)
        q = Popen(getinfo, stdin=p.stdout, stdout=f)
    else:
        p = Popen(pval.split(), stdin=PIPE, stdout=f)
        q = p
    p.stdin.write(np2bytes(ar[-1::-1, -1::-1].T))
    q.communicate()
    f.close()
    return imgf


def carray2hdr(ar, imgf, header=None):
    """write color channel np.array (3, x, y) to hdr image format

    Parameters
    ----------
    ar: np.array
    imgf: file path to right
    header: list of header lines to append to image header

    Returns
    -------

    """
    f = open(imgf, 'wb')
    pval = f'pvalue -r -h -H -df -o -y {ar.shape[-1]} +x {ar.shape[-2]}'
    if header is not None:
        hdr = "' '".join(header)
        getinfo = shlex.split(f"getinfo -a '{hdr}'")
        p = Popen(pval.split(), stdin=PIPE, stdout=PIPE)
        q = Popen(getinfo, stdin=p.stdout, stdout=f)
    else:
        p = Popen(pval.split(), stdin=PIPE, stdout=f)
        q = p
    p.stdin.write(np2bytes(ar.T[-1::-1, -1::-1, :]))
    q.communicate()
    f.close()
    return imgf


def hdr2array(imgf):
    """read np.array from hdr image

    Parameters
    ----------
    imgf: file path of image

    Returns
    -------
    ar: np.array

    """
    pval = f'pvalue -b -h -df -o {imgf}'
    p = Popen(pval.split(), stdout=PIPE)
    shape = p.stdout.readline().strip().split()
    shape = (int(shape[-3]), int(shape[-1]))
    return bytes2np(p.stdout.read(), shape)[-1::-1, -1::-1]


def rgbe2lum(rgbe):
    """
    convert from Radiance hdr rgbe 4-byte data format to floating point
    luminance.

    Parameters
    ----------
    rgbe: np.array
        r,g,b,e unsigned integers according to:
        http://radsite.lbl.gov/radiance/refer/filefmts.pdf

    Returns
    -------
    lum: luminance in cd/m^2
    """
    v = np.power(2., rgbe[:, 3] - 128).reshape(-1, 1) / 256
    lum = np.where(rgbe[:, 0:3] == 0, 0, (rgbe[:, 0:3] + 0.5) * v)
    # luminance = 179 * (0.265*R + 0.670*G + 0.065*B)
    return np.einsum('ij,j', lum, [47.435, 119.93, 11.635])



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import os
import sys
from concurrent.futures import ProcessPoolExecutor, as_completed
from datetime import datetime, timezone
import subprocess

import numpy as np
import raytraverse
from raytraverse import translate, io, skycalc, metric
from raytraverse.mapper import ViewMapper
from raytraverse.scene import SkyInfo


class Integrator(object):
    """class to generate outputs from skyfield sunfield and sky conditions

    This class provides an interface to:

        1. generate sky data using the perez
        2. combine lightfield data for sky and sun (handling sparsely populated
           sun data)
        3. apply sky data to the lightfield queries
        4. output luminance maps and photometric quantities and visual
           comfort metrics

    Parameters
    ----------
    skyfield: raytraverse.lightfield.SCBinField
        class containing sky data
    sunfield: raytraverse.lightfield.SunField
        class containing sun data (should share its scene parameter with the
        given skyfield
    wea: str, np.array, optional
        path to epw, wea, or .npy file or np.array, if loc not set attempts to
        extract location data (if needed). The Integrator does not need to be
        initialized with weather data but for convinience can be. If skydata
        is not set then the optional parameter of get_sky_matrix() is required
        or an Exception will be raised.
    loc: (float, float, int), optional
        location data given as lat, lon, mer with + west of prime meridian
        overrides location data in wea (but not in sunfield)
    skyro: float, optional
        angle in degrees counter-clockwise to rotate sky
        (to correct model north, equivalent to clockwise rotation of scene)
        does not override rotation in SunField)
    """

    def __init__(self, skyfield, sunfield=None, wea=None, loc=None, skyro=0.0,
                 **kwargs):
        #: raytraverse.scene.Scene
        self.scene = skyfield.scene
        try:
            #: raytraverse.sunsetter.SunSetter
            self.suns = sunfield.suns
        except AttributeError:
            self.suns = None
        try:
            #: raytraverse.scene.SkyInfo
            self.sky = self.suns.sky
        except AttributeError:
            if loc is None and wea is not None:
                try:
                    loc = skycalc.get_loc_epw(wea)
                except ValueError:
                    pass
            self.sky = SkyInfo(loc, skyro=skyro)
        #: raytraverse.lightfield.SunField
        self.sunfield = sunfield
        #: raytraverse.lightfield.SCBinField
        self.skyfield = skyfield
        if wea is not None:
            self.skydata = wea
        else:
            try:
                self.skydata = f'{self.scene.outdir}/skydat.txt'
            except OSError:
                self._skydata = None

    @property
    def skydata(self):
        """sky data formatted as dx, dy, dz, dirnorm, diffhoriz

        :getter: Returns this scene's skydata
        :setter: Sets this scene's skydata from file path or
        :type: np.array
        """
        return self._skydata

    @skydata.setter
    def skydata(self, wea):
        self._skydata = self.format_skydata(wea)
        np.savetxt(f'{self.scene.outdir}/skydat.txt', self._skydata)

    def format_skydata(self, dat):
        """process dat argument as skydata

        Parameters
        ----------
        dat: str, np.array
            This method takes either a file path or np.array. File path can
            point to a wea, epw, or .npy file. Loaded array must be one of the
            following:
            - 4 col: alt, az, dir, diff
            - 5 col: dx, dy, dz, dir, diff
            - 5 col: m, d, h, dir, diff'

        Returns
        -------
        np.array
            dx, dy, dz, dir, diff
        """
        loc = self.sky.loc
        if hasattr(dat, 'shape'):
            skydat = dat
        else:
            try:
                skydat = np.loadtxt(dat)
            except ValueError:
                if self.sky.loc is None:
                    loc = skycalc.get_loc_epw(dat)
                skydat = skycalc.read_epw(dat)
        if skydat.shape[1] == 4:
            xyz = translate.aa2xyz(skydat[:, 0:2])
            skydat = np.hstack((xyz, skydat[:, 2:]))
        elif skydat.shape[1] == 5:
            if np.max(skydat[:, 2]) > 2:
                if self.sky.loc is None:
                    raise ValueError("cannot parse wea data without a Location")
                times = skycalc.row_2_datetime64(skydat[:, 0:3])
                xyz = skycalc.sunpos_xyz(times, *loc, ro=self.sky.skyro)
                skydat = np.hstack((xyz, skydat[:, 3:]))
        else:
            raise ValueError('input data should be one of the following:'
                             '\n4 col: alt, az, dir, diff'
                             '\n5 col: dx, dy, dz, dir, diff'
                             '\n5 col: m, d, h, dir, diff')
        return skydat

    def get_sky_mtx(self, skydata=None):
        """generate sky, grnd and sun coefficients from sky data using perez

        Parameters
        ----------
        skydata: str, np.array. optional
            if None, uses object skydata (will raise error if unassigned)
            see format_skydata() for argument specifics.

        Returns
        -------
        smtx: np.array
            shape (len(skydata), skyres**2) coefficients for each sky patch
            each row is a timestep, timesteps where a sun exists exclude the
            sun coefficient, otherwise the patch enclosing the sun position
            contains the energy of the sun
        grnd: np.array
            shape (len(skydata),) coefficients for ground at each timestep
        sun: np.array
            shape (len(skydata), 5) sun position (index 0,1,2) and coefficients
            for sun at each timestep assuming the true solid angle of the sun
            (index 3) and the weighted value for the sky patch (index 4).
        si: np.array
             shape (len(skydata), 2) index for self.suns.suns pointing to
             correct proxy source (col 0) and sunbin for patch mapping (col 1)
        daysteps: np.array
            shape (len(skydata),) boolean array masking timesteps when sun is
            below horizon
        """
        if skydata is None:
            skydata = self.skydata
        else:
            skydata = self.format_skydata(skydata)
        daysteps = skydata[:, 2] > 0
        sxyz = skydata[daysteps, 0:3]
        if self.suns is not None:
            si = self.suns.proxy_src(sxyz, tol=self.suns.sunres)
            # nosun = np.arange(hassun.size)[np.logical_not(hassun)]
        else:
            # nosun = np.arange(sxyz.shape[0])
            # hassun = 0
            si = np.zeros(sxyz.shape[0])
        sunuv = translate.xyz2uv(sxyz, flipu=False)
        sunbins = translate.uv2bin(sunuv, self.scene.skyres)
        dirdif = skydata[daysteps, 3:]
        smtx, grnd, sun = skycalc.sky_mtx(sxyz, dirdif, self.scene.skyres)
        # ratio between actual solar disc and patch
        omegar = np.square(0.2665 * np.pi * self.scene.skyres / 180) * .5
        plum = sun[:, -1] * omegar
        sun = np.hstack((sun, plum[:, None]))
        si = np.stack((si, sunbins)).T
        return smtx, grnd, sun, si, daysteps

    def header(self):
        """generate image header string"""
        octe = f"{self.scene.outdir}/scene.oct"
        hdr = subprocess.run(f'getinfo {octe}'.split(), capture_output=True,
                             text=True)
        hdr = [i.strip() for i in hdr.stdout.split('\n')]
        hdr = [i for i in hdr if i[0:5] == 'oconv']
        tf = "%Y:%m:%d %H:%M:%S"
        hdr.append("CAPDATE= " + datetime.now().strftime(tf))
        hdr.append("GMT= " + datetime.now(timezone.utc).strftime(tf))
        radversion = subprocess.run('rpict -version'.split(),
                                    capture_output=True,
                                    text=True)
        hdr.append(f"SOFTWARE= {radversion.stdout}")
        lastmod = os.path.getmtime(os.path.dirname(raytraverse.__file__))
        tf = "%a %b %d %H:%M:%S %Z %Y"
        lm = datetime.fromtimestamp(lastmod, timezone.utc).strftime(tf)
        hdr.append(
            f"SOFTWARE= RAYTRAVERSE {raytraverse.__version__} lastmod {lm}")
        try:
            hdr.append("LOCATION= lat: {} lon: {} tz: {}".format(*self.sky.loc))
        except TypeError:
            pass
        return hdr

    def _hdr(self, res, vm, keymap, pi, suni, sun, pdirs, si, vstr, outf,
             interp, mask, skyv):
        img = np.zeros((res*vm.aspect, res))
        if keymap[pi, suni[0]] and sun[-1] > 0:
            sun = sun[0:4]
            psi = (pi, suni[0])
            j, e = self.sunfield.query_ray(psi, pdirs[mask],
                                           interp=interp)
            self.sunfield.add_to_img(img, mask, psi, j, e, sun[-1], vm)
        else:
            skyv[suni[1]] += sun[4]
        self.skyfield.add_to_img(img, mask, pi, *si, coefs=skyv)
        io.array2hdr(img, outf, self.header() + [vstr])
        return outf

    def hdr(self, pts, smtx, grnd=None, suns=None, suni=None,
            daysteps=None, vname='view', viewangle=180.0, res=400, interp=1):
        """

        Parameters
        ----------
        pts: np.array
            points and view dirs shape (N, 6)
        smtx: np.array
            shape (len(skydata), skyres**2) coefficients for each sky patch
            each row is a timestep, timesteps where a sun exists exclude the
            sun coefficient, otherwise the patch enclosing the sun position
            contains the energy of the sun
        grnd: np.array
            shape (len(skydata),) coefficients for ground at each timestep
        suns: np.array
            shape (len(skydata), 5) sun position (index 0,1,2) and coefficients
            for sun at each timestep assuming the true solid angle of the sun
            (index 3) and the weighted value for the sky patch (index 4).
        suni: np.array
             shape (len(skydata), 2) index for self.suns.suns pointing to
             correct proxy source (col 0) and sunbin for patch mapping (col 1)
        daysteps: np.array
            shape (len(skydata),) boolean array masking timesteps when sun is
            below horizon
        vname: str
            view name for output file
        viewangle: float, optional
            degree opening of view cone
        res: int, optional
            image resolution
        interp: int, optional
            number of nearest points to interpolate between. 1 will resemble
            voronoi patches

        Returns
        -------

        """
        npts = self.scene.area.npts
        perrs, pis = self.scene.area.pt_kd.query(pts[:, 0:3])
        if self.suns:
            keymap = self.sunfield.keymap()
        else:
            keymap = np.full((npts, 1), False)
        with ProcessPoolExecutor() as exc:
            fu = []
            for pj, (pi, vdir) in enumerate(zip(pis, pts[:, 3:6])):
                vm = ViewMapper(viewangle=viewangle, dxyz=vdir, name=vname)
                vp = self.scene.area.idx2pt([pi])[0]
                vstr = ('VIEW= -vta -vv {0} -vh {0} -vd {1} {2} {3}'
                        ' -vp {4} {5} {6}'.format(viewangle, *vdir, *vp))
                pdirs = vm.pixelrays(res)
                mask = vm.in_view(pdirs)
                si = self.skyfield.query_ray(pi, pdirs[mask], interp=interp)
                for sj, skyv in enumerate(smtx):
                    outf = (f"{self.scene.outdir}_{vname}_{pj:04d}_{sj:04d}"
                            ".hdr")
                    fu.append(exc.submit(self._hdr, res, vm, keymap, pi,
                                         suni[sj], suns[sj], pdirs, si, vstr,
                                         outf, interp, mask, skyv))
            for future in as_completed(fu):
                print(future.result(), file=sys.stderr)

    def metric(self, pts, smtx, grnd=None, suns=None, suni=None,
               daysteps=None, metricfuncs=(metric.illum,), **kwargs):
        """calculate luminance based metrics for given sensors and skyvecs

        Parameters
        ----------
        pts: np.array
            points and view dirs shape (N, 6)
        smtx: np.array
            shape (len(skydata), skyres^2) coefficients for each sky patch
            each row is a timestep, timesteps where a sun exists exclude the
            sun coefficient, otherwise the patch enclosing the sun position
            contains the energy of the sun
        grnd: np.array
            shape (len(skydata),) coefficients for ground at each timestep
        suns: np.array
            shape (len(skydata), 5) sun position (index 0,1,2) and coefficients
            for sun at each timestep assuming the true solid angle of the sun
            (index 3) and the weighted value for the sky patch (index 4).
        suni: np.array
             shape (len(skydata), 2) index for self.suns.suns pointing to
             correct proxy source (col 0) and sunbin for patch mapping (col 1)
        daysteps: np.array
            shape (len(skydata),) boolean array masking timesteps when sun is
            below horizon
        metricfuncs: tuple
            list of metric functions to apply, with the call signature:
            f(view, rays, omegas, lum, kwargs)

        Returns
        -------
        metrics: np.array
            metrics at each point/direction and sky weighting axes ordered as:
            (view, points, skys, metrics) an additional column on the final axis
            is 0 when only the skyfield is used, 1 when sun reflections exists
            and 2 when a view to the sun also exists.
        """
        npts = self.scene.area.npts
        perrs, pis = self.scene.area.pt_kd.query(pts[:, 0:3])
        if self.suns:
            keymap = self.sunfield.keymap()
        else:
            keymap = np.full((npts, 1), False)
        pmetrics = []
        for pi, vdir in zip(pis, pts[:, 3:6]):
            # get sky vectors
            skyidx = self.skyfield.query_ball(pi, [vdir])
            smetrics = []
            vm = ViewMapper(vdir, viewangle=180)
            for skyv, s, (si, sb) in zip(smtx, suns, suni):
                psi = (pi, si)
                if keymap[pi, si] and s[-2] > 0:
                    skyvec = skyv
                    sunidx = self.sunfield.query_ball(psi, [vdir])
                    sunlm = np.squeeze(self.sunfield.apply_coef(psi, s[-2]))
                else:
                    skyvec = np.copy(skyv)
                    skyvec[sb] += s[-1]
                    sunidx = False
                    sunlm = False
                skylm = np.squeeze(self.skyfield.apply_coef(pi, skyvec))
                skyrays = self.skyfield.vec[pi][skyidx[0]]
                osky = np.squeeze(self.skyfield.omega[pi][skyidx[0]])
                lmsky = skylm[skyidx[0]]
                sunvalue = 0
                mstack = []
                skyweight = 1
                if sunidx:
                    sunvalue += 1
                    sunrays = self.sunfield.vec[psi][sunidx[0]]
                    osun = np.squeeze(self.sunfield.omega[psi][sunidx[0]])
                    lmsun = sunlm[sunidx[0]]

                    # TODO cross match samples
                    # currently implements a nearest neighbor and then
                    # averages the two results (using the skyweight term,
                    # this is subject to some error a better approach is
                    # to query the ray for neighbors with a radius derived
                    # from omega and average in the case of no result
                    # (when omega_other is larger) use the nearest neighbor.
                    # This will likely need to be implemented in c++ to be
                    # practical, and alternative is to interpolate samples
                    # like hdr image generation and then naively add.
                    sunsky, err = self.sunfield.query_ray(psi, skyrays)
                    skysun, err = self.skyfield.query_ray(pi, sunrays)
                    lmsun += skylm[skysun]
                    lmsky += sunlm[sunsky]
                    skyweight = .5

                    fsun = [f(vm, sunrays, osun, lmsun, **kwargs) * skyweight
                            for f in metricfuncs]
                    mstack.append(fsun)
                fsky = [f(vm, skyrays, osky, lmsky, **kwargs)*skyweight
                        for f in metricfuncs]
                mstack.append(fsky)
                try:
                    fsv = self.sunfield.view.metric(psi, vm, s, metricfuncs,
                                                    **kwargs)
                except (AttributeError, ValueError):
                    pass
                else:
                    sunvalue += 1
                    mstack.append(fsv)
                smetrics.append(np.concatenate((np.sum(mstack, 0), [sunvalue])))
            pmetrics.append(smetrics)
        # loop returns shape: (points, skys, metrics)
        # transpose reorganizes as: (view, points, skys, metrics)
        return np.array(pmetrics)



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import sys
import os
from raytraverse import io
from raytraverse.renderer.renderer import Renderer


class RadianceRenderer(Renderer):
    """Virtual class for wrapping c++ Radiance renderer executable classes"""

    returnbytes = False

    def __new__(cls, rayargs=None, scene=None, nproc=None, iot="ff"):
        cls.instance = cls.Engine.get_instance()
        return super().__new__(cls, rayargs=rayargs, scene=scene, nproc=nproc,
                               iot=iot)

    @classmethod
    def update_param(cls, args, nproc=None, iot="ff"):
        cls.returnbytes = iot[-1] != "a"
        if nproc is None:
            nproc = os.cpu_count()
        cls.initialized = cls._set_args(args, iot, nproc)
        cls.instance.initialize(cls.initialized)

    @classmethod
    def initialize(cls, args, scene, nproc=None, iot="ff"):
        if cls.instance is None:
            cls.instance = cls.Engine.get_instance()
        if args is not None:
            firstload = not cls.initialized
            cls.update_param(args, nproc, iot)
            if firstload:
                cls.instance.load_scene(scene)
                # TODO: populate header
                cls.header = ""

    @classmethod
    def call(cls, rayfile, store=True, outf=None):
        if not cls.initialized:
            raise ValueError(f'{cls.__name__} instance not initialized')
        with io.CaptureStdOut(cls.returnbytes, store, outf) as capture:
            cls.instance.call(rayfile)
        return capture.stdout

    @classmethod
    def reset(cls):
        cls.instance.reset()
        super().reset()

    @classmethod
    def reset_instance(cls):
        cls.instance.reset_instance()
        super().reset_instance()

